#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        montecarlo.py
# Purpose:     Classes to automate Monte-Carlo simulations
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     10-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Union, Optional, Dict, Callable, Type

from ...editor.base_editor import BaseEditor, scan_eng
from .sim_analysis import SimAnalysis, AnyRunner, ProcessCallback
from enum import Enum


class DeviationType(Enum):
    """Enum to define the type of deviation"""
    tolerance = 'tolerance'
    minmax = 'minmax'
    none = 'none'


@dataclass
class ComponentDeviation:
    """Class to store the deviation of a component"""
    max_val: float
    min_val: float = 0.0
    typ: DeviationType = DeviationType.tolerance
    distribution: str = 'uniform'

    @classmethod
    def from_tolerance(cls, tolerance: float, distribution: str = 'uniform'):
        return cls(tolerance, -tolerance, DeviationType.tolerance, distribution)

    @classmethod
    def from_min_max(cls, min_val: float, max_val: float, distribution: str = 'uniform'):
        return cls(min_val, max_val, DeviationType.minmax, distribution)

    @classmethod
    def none(cls):
        return cls(0.0, 0.0, DeviationType.none)


class ToleranceDeviations(SimAnalysis, ABC):
    """Class to automate Monte-Carlo simulations"""
    devices_with_deviation_allowed = ('R', 'C', 'L', 'V', 'I')

    def __init__(self, circuit_file: Union[str, BaseEditor], runner: Optional[AnyRunner] = None):
        super().__init__(circuit_file, runner)
        self.default_tolerance = {prefix: ComponentDeviation.none() for prefix in self.devices_with_deviation_allowed}
        self.device_deviations: Dict[str, ComponentDeviation] = {}
        self.parameter_deviations: Dict[str, ComponentDeviation] = {}
        self.testbench_prepared = False
        self.testbench_executed = False
        self.num_runs = 0
        self.simulation_results = {}

    def reset_tolerances(self):
        """
        Clears all the settings for the simulation
        """
        self.device_deviations.clear()
        self.parameter_deviations.clear()
        self.testbench_prepared = False
        self.num_runs = 0

    def clear_simulation_data(self):
        """Clears the data from the simulations"""
        super().clear_simulation_data()
        self.simulation_results.clear()

    def set_tolerance(self, ref: str, new_tolerance: float, distribution: str = 'uniform'):
        """
        Sets the tolerance for a given component. If only the prefix is given, the tolerance is set for all.
        The valid prefixes that can be used are: R, C, L, V, I
        """
        if ref in self.devices_with_deviation_allowed:  # Only the prefix is given
            self.default_tolerance[ref] = ComponentDeviation.from_tolerance(new_tolerance, distribution)
        else:
            if ref in self.editor.get_components(ref[0]):
                self.device_deviations[ref] = ComponentDeviation.from_tolerance(new_tolerance, distribution)

    def set_tolerances(self, new_tolerances: dict, distribution: str = 'uniform'):
        """
        Sets the tolerances for a set of components. The dictionary keys are the references and the values are the
        tolerances. If only the prefix is given, the tolerance is set for all components with that prefix. See
        set_tolerance method.
        """
        for ref, tol in new_tolerances.items():
            self.set_tolerance(ref, tol, distribution)

    def set_deviation(self, ref: str, min_val, max_val: float, distribution: str = 'uniform'):
        """
        Sets the deviation for a given component. This establishes a min and max value for the component.
        Optionally a distribution can be specified. The valid distributions are: uniform or normal (gaussian).
        """
        self.device_deviations[ref] = ComponentDeviation.from_min_max(min_val, max_val, distribution)

    def get_components(self, prefix: str):
        if prefix == '*':
            return (cmp for cmp in self.editor.get_components() if cmp[0] in self.devices_with_deviation_allowed)
        return self.editor.get_components(prefix)

    def get_component_value_deviation_type(self, ref: str) -> (float, ComponentDeviation):
        if ref[0] not in self.devices_with_deviation_allowed:
            raise ValueError("The reference must be a valid component type")
        value = self.editor.get_component_value(ref)
        if len(value) == 0:  # This covers empty strings
            return value, ComponentDeviation.none()
        # The value needs to be able to be computed, otherwise it can't be used
        try:
            value = scan_eng(value)
        except ValueError:
            if value.startswith('{') and value.endswith('}'):
                # This is still acceptable as the value could be computed.
                # but we need to get rid of the outer {}
                value = value[1:-1]
            else:
                return value, ComponentDeviation.none()
        if ref in self.device_deviations:
            return value, self.device_deviations[ref]
        elif ref[0] in self.default_tolerance:
            return value, self.default_tolerance[ref[0]]
        else:
            return value, ComponentDeviation.none()

    def set_parameter_deviation(self, ref: str,  min_val, max_val: float, distribution: str = 'uniform'):
        self.parameter_deviations[ref] = ComponentDeviation.from_min_max(min_val, max_val, distribution)

    def get_parameter_value_deviation_type(self, param: str) -> (float, ComponentDeviation):
        value = self.editor.get_parameter(param)
        return value, self.parameter_deviations[param]

    def save_netlist(self, filename: str):
        if self.testbench_prepared is False:
            self.prepare_testbench()
        super().save_netlist(filename)

    def _reset_netlist(self):
        super()._reset_netlist()
        self.testbench_prepared = False

    @abstractmethod
    def prepare_testbench(self, **kwargs):
        ...

    def run_testbench(self, *,
            max_runs_per_sim: int = 512,
            wait_resource: bool = True,
            callback: Union[Type[ProcessCallback], Callable] = None,
            callback_args: Union[tuple, dict] = None,
            switches=None,
            timeout: float = None,
            run_filename: str = None):
        """
        Runs the simulations.
        :param max_runs_per_sim: Maximum number of runs per simulation. If the number of runs is higher than this number,
        the simulation is split in multiple runs.
        :param wait_resource: If True, the simulation will wait for the resource to be available. If False, the simulation
        will be queued and the method will return immediately.
        :param callback: A callback function to be called when the simulation is completed. The callback function must
        accept a single argument, which is the simulation object.
        :param callback_args: A tuple or dictionary with the arguments to be passed to the callback function.
        :param switches: A dictionary with the switches to be passed to the simulator.
        :param timeout: A timeout in seconds. If the simulation is not completed in this time, it will be aborted.
        :param run_filename: The name of the file to be used for the simulation. If None, a temporary file will be used.
        :return: The callback returns of every batch if a callback function is given. Otherwise, None.
        """
        super()._reset_netlist()
        self.clear_simulation_data()
        self.play_instructions()
        self.prepare_testbench()
        self.editor.remove_instruction(".step param run -1 %d 1" % self.num_runs)  # Needs to remove this instruction
        for sim_no in range(-1, self.num_runs, max_runs_per_sim):
            last_no = sim_no + max_runs_per_sim - 1
            if last_no > self.num_runs:
                last_no = self.num_runs
            if sim_no >= last_no:
                break
            run_stepping = ".step param run {} {} 1".format(sim_no, last_no)
            self.editor.add_instruction(run_stepping)
            sim = self.runner.run(self.editor, wait_resource=wait_resource, callback=callback,
                                  callback_args=callback_args, switches=switches, timeout=timeout,
                                  run_filename=run_filename)
            self.simulations.append(sim)
            self.editor.remove_instruction(run_stepping)
        self.runner.wait_completion()
        if callback is not None:
            return (sim.get_results() if sim is not None else None for sim in self.simulations)
        self.testbench_executed = True
        return None

    @abstractmethod
    def run_analysis(self):
        ...



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        raw_write.py
# Purpose:     Create RAW Files
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     16-10-2021
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

"""
This module generates RAW Files from user data.
It can be used to combine RAW files generated by different Simulation Runs
"""
from typing import Union
from time import strftime

from .raw_read import RawRead
from .raw_classes import DataSet
from numpy import array, float32, zeros


class Trace(DataSet):
    """Helper class representing a trace. This class is based on DataSet, therefore, it doesn't support STEPPED data.
    :param name: name of the trace being created
    :type name: str
    :param whattype: time, frequency, voltage or current
    :type whattype: str
    :param data: data for the data write
    :type data: list or numpy.array
    :param numerical_type: real or complex
    :type numerical_type: str
    """

    def __init__(self, name, data, whattype='voltage', numerical_type=''):
        if name == 'time':
            whattype = 'time'
        elif name == 'frequency':
            whattype = 'frequency'
        if numerical_type == '':
            if name == 'time':
                numerical_type = 'double'
            elif name == 'frequency':
                raise AssertionError("For frequency plots, please specify the numerical_type:\n"
                                     "Use:\n"
                                     "   * numerical_type='complex' for .AC analysis\n"
                                     "   * numerical_type='double' for .NOISE analysys")
            elif isinstance(data[0], float32) or isinstance(data[0], float):
                numerical_type = 'real'
            elif isinstance(data[0], complex):
                numerical_type = 'complex'
            else:
                raise NotImplementedError

        DataSet.__init__(self, name, whattype, len(data), numerical_type=numerical_type)
        if isinstance(data, (list, tuple)):
            self.data = array(data, dtype=self.data.dtype)
        else:
            self.data[:] = data[:]  # This way the dtype is kept


class RawWrite(object):
    """
    This class represents the RAW data file being generated. Contrary to the RawRead this class doesn't support stepped
    data.

    """

    def __init__(self, plot_name=None, fastacces=True, numtype='auto', encoding='utf_16_le'):
        self._traces = list()
        self.flag_numtype = numtype
        self.flag_forward = False
        self.flag_log = False
        self.flag_stepped = False
        self.flag_fastaccess = fastacces
        self.plot_name = plot_name
        self.offset = 0.0
        self.encoding = encoding
        self._imported_data = []
        self._new_axis = None

    def _str_flags(self):
        flags = [self.flag_numtype]
        if self.flag_forward:
            flags.append('forward')
        if self.flag_log:
            flags.append('log')
        if self.flag_stepped:
            flags.append('stepped')
        if self.flag_fastaccess:
            flags.append('fastaccess')
        return ' '.join(flags)

    def add_trace(self, trace: Trace):
        """
        Adds a trace to the RAW file. The trace needs to have the same size as trace 0 ('time', 'frequency', etc..)
        The first trace added defines the X-Axis and therefore the type of RAW file being generated. If no plot name
        was defined, it will automatically assign a name.
        :param trace: Needs to be of the
        :type trace:
        :return: Nothing
        :rtype: None
        """
        assert isinstance(trace, Trace), "The trace needs to be of the type ""Trace"""
        if len(self._traces) == 0:
            if trace.whattype == 'time':
                self.plot_name = self.plot_name or 'Transient Analysis'
                flag_numtype = 'real'
            elif trace.whattype == 'frequency':
                if (trace.numerical_type != 'complex' and self.flag_numtype != 'complex') or 'Noise' in self.plot_name:
                    self.plot_name = self.plot_name or 'Noise Spectral Density - (V/Hz½ or A/Hz½)'
                    flag_numtype = 'real'
                else:
                    self.plot_name = self.plot_name or 'AC Analysis'
                    flag_numtype = 'complex'
            elif trace.whattype in ('voltage', 'current'):
                self.plot_name = self.plot_name or 'DC transfer characteristic'
                flag_numtype = 'real'
            elif trace.whattype == 'param':
                self.plot_name = self.plot_name or 'Operating Point'
                flag_numtype = 'real'
            else:
                raise ValueError("First Trace needs to be either 'time', 'frequency', 'param', 'voltage' or '...'")

            if self.flag_numtype == 'auto':
                self.flag_numtype = flag_numtype
        else:
            if len(self._traces[0]) != len(trace):
                raise IndexError("The trace needs to be the same size as trace 0")
        self._traces.append(trace)

    def save(self, filename: str):
        """
        Saves the RAW file into a file. The file format is always binary. Text based RAW output format is not supported
        in this version.
        :param filename: filename to where the RAW file is going to be written. Make sure that the extension of the
        file is .RAW.

        :type filename: str or pathlib.Path
        :return: Nothing
        :rtype: None
        """
        if len(self._imported_data):
            self._consolidate()
        f = open(filename, 'wb')
        f.write("Title: * spicelib RawWrite\n".encode(self.encoding))
        f.write("Date: {}\n".format(strftime("%a %b %d %H:%M:%S %Y")).encode(self.encoding))
        f.write("Plotname: {}\n".format(self.plot_name).encode(self.encoding))
        f.write("Flags: {}\n".format(self._str_flags()).encode(self.encoding))
        f.write("No. Variables: {}\n".format(len(self._traces)).encode(self.encoding))
        f.write("No. Points: {:12}\n".format(len(self._traces[0])).encode(self.encoding))
        f.write("Offset:   {:.16e}\n".format(self.offset).encode(self.encoding))
        f.write("Command: Linear Technology Corporation LTspice XVII\n".encode(self.encoding))
        # f.write("Backannotation: \n".encode(self.encoding))
        f.write("Variables:\n".encode(self.encoding))
        for i, trace in enumerate(self._traces):
            f.write("\t{0}\t{1}\t{2}\n".format(i, trace.name, trace.whattype).encode(self.encoding))
        total_bytes = 0
        f.write("Binary:\n".encode(self.encoding))
        if self.flag_fastaccess and self.flag_numtype != 'complex':  # Don't know why, but complex RAW files aren't
            # converted to FastAccess
            for trace in self._traces:
                f.write(trace.data.tobytes())
        else:
            fmts = {trace: tobytes_for_trace(trace) for trace in self._traces}
            for i in range(len(self._traces[0])):
                for trace in self._traces:
                    total_bytes += f.write(fmts[trace](trace.data[i]))
        f.close()

    @staticmethod
    def _rename_netlabel(name, **kwargs) -> str:
        """Renames a trace name making sure that the V() or I() containers are left intact."""
        # Make the rename as requested
        if 'rename_format' in kwargs:
            if name.endswith(')') and name.startswith('V(') or name.startswith('I('):
                new_name = name[:2] + kwargs['rename_format'].format(name[2:-1], **kwargs) + name[-1]
            else:
                new_name = kwargs['rename_format'].format(name, **kwargs)
            return new_name
        else:
            return name

    def _name_exists(self, name: str) -> bool:
        # first check whether it is a duplicate
        for trace in self._traces:
            if trace.name == name:
                return True
        return False

    def add_traces_from_raw(self, other: RawRead, trace_filter: Union[list, tuple, str], **kwargs):
        """ *(Not fully implemented)*

        Merge two RawWrite classes together resulting in a new instance
        :param other: an instance of the RawRead class where the traces are going to be copied from.
        :type other: RawRead
        :param trace_filter: A list of signals that should be imported into the new file
        :type trace_filter: list, Tuple, or just a string for one trace

        :keyword force_axis_alignment: If two raw files don't have the same axis, an attempt is made to sync the two

        :keyword admissible_error: maximum error allowed in the sync between the two axis

        :keyword rename_format: when adding traces with the same name, it is possible to define a rename format.
            For example, if there are two traces named N001 in order to avoid duplicate names the rename format can be
            defined as "{}_{kwarg_name} where kwarg_name is passed as a keyword argument of this function. If just one
            trace is being added, this can be used to simply give the new name.

        :keyword step: by default only step 0 is added from the second raw. It is possible to add other steps, by
            using this keyword parameter. This is useful when we want to "flatten" the multiple step runs into the same
            view.

        :keyword: minimum_timestep: This parameter forces the two axis to sync using a minimum time step. That is, all
            time increments that are less than this parameter will be suppressed.

        :returns: Nothing
        """
        force_axis_alignment = kwargs.get('force_axis_alignment', False)
        admissible_error = kwargs.get('admissible_error', 1e-11)
        from_step = kwargs.get('step', 0)
        minimum_timestep = kwargs.get('minimum_timestep', 0.0)
        if isinstance(trace_filter, str):
            trace_filter = [trace_filter]

        other_flags = other.get_raw_property('Flags').split(' ')
        for flag in other_flags:
            if flag in ('real', 'complex'):
                other_flag_num_type = flag
                break
        else:
            other_flag_num_type = 'real'

        if len(self._traces):  # there are already traces
            if self.flag_numtype != other_flag_num_type:
                raise ValueError("The two instances should have the same type:\n"
                                 f"Source is {other_flag_num_type} and Destination is {self.flag_numtype}")
            if self._traces[0].whattype != other.get_trace(0).whattype:
                raise ValueError("The two instances should have the same axis type:\n"
                                 f"Source is {other.get_trace(0).whattype} and Destination is {self._traces[0].whattype}")
            if len(trace_filter) == 0:
                return  # There is nothing to add stop here

        else:  # No traces are present
            # if no X axis is present, copy from the first one
            self.flag_numtype = other_flag_num_type
            self.flag_log = 'log' in other_flags
            self.flag_forward = 'forward' in other_flags
            self.plot_name = other.get_raw_property('Plotname')
            oaxis = other.get_trace(0)
            new_axis = Trace(oaxis.name, other.get_axis(from_step), oaxis.whattype, oaxis.numerical_type)
            self._traces.append(new_axis)
            force_axis_alignment = False

        if force_axis_alignment or minimum_timestep > 0.0:
            if self._new_axis:
                my_axis = self._new_axis
            else:
                my_axis = self._traces[0].get_wave()
            other_axis = other.get_axis(from_step)
            new_axis = []
            if minimum_timestep > 0.0:
                raise NotImplementedError
            else:
                i = 0  # incomming data counter
                e = 0  # existing data counter

                while e < len(my_axis)-1 and i < len(other_axis)-1:
                    error = other_axis[i] - my_axis[e]
                    if abs(error) < admissible_error:
                        new_axis.append(my_axis[e])
                        i += 1
                        e += 1
                    elif error < 0:
                        # Other axis is smaller
                        new_axis.append(other_axis[i])
                        i += 1
                    else:
                        new_axis.append(my_axis[e])
                        e += 1
                # Creating the New Axis
                self._new_axis = new_axis

                for trace_name in trace_filter:
                    imported_trace = other.get_trace(trace_name)
                    new_name = self._rename_netlabel(trace_name, **kwargs)
                    imported_trace.name = new_name
                    self._imported_data.append(imported_trace)
        else:
            assert len(self._traces[0]) == len(other.get_axis(from_step)), \
                "The two instances should have the same size. To avoid this use force_axis_alignment=True option"
            for trace_name in trace_filter:
                trace = other.get_trace(trace_name)
                new_name = self._rename_netlabel(trace_name, **kwargs)
                data = trace.get_wave(from_step)
                self._traces.append(Trace(new_name, data, trace.whattype, numerical_type=trace.numerical_type))

    @staticmethod
    def _interpolate(trace_data, trace_axis, new_axis: array):
        new_data = zeros(len(new_axis), dtype=trace_data.dtype)
        new_data[0] = trace_data[0]

        slope = (trace_data[1] - trace_data[0])/(trace_axis[1] - trace_axis[0])
        i = 1
        for j, t in enumerate(new_axis):
            while trace_axis[i] < t:
                i += 1
                slope = (trace_data[i] - trace_data[i-1])/(trace_axis[i] - trace_axis[i-1])
            new_data[j] = trace_data[i-1] + slope * (t - trace_axis[i-1])
        return new_data

    def _consolidate(self):
        if self._new_axis and self._imported_data:
            new_axis = self._new_axis
            axis_length = len(new_axis)
            old_axis = self._traces[0]
            if axis_length != len(old_axis):
                my_axis = old_axis.data
                for trace in self._traces[1:]:
                    trace.data = self._interpolate(trace.data, my_axis, new_axis)
            for imported_trace in self._imported_data:
                new_trace = Trace(imported_trace.name,
                                  self._interpolate(imported_trace.get_wave(), imported_trace.axis.get_wave(), new_axis),
                                  imported_trace.whattype, imported_trace.numerical_type)
                self._traces.append(new_trace)
            self._traces[0] = Trace(old_axis.name, new_axis,
                                    old_axis.whattype, old_axis.numerical_type)  # Replaces with the new axis
            self._new_axis = None
            self._imported_data.clear()

    def get_trace(self, trace_ref):
        """
        Retrieves the trace with the requested name (trace_ref).

        :param trace_ref: Name of the trace
        :type trace_ref: str
        :return: An object containing the requested trace
        :rtype: DataSet subclass
        """
        if isinstance(trace_ref, str):
            for trace in self._traces:
                if trace_ref == trace.name:
                    # assert isinstance(trace, DataSet)
                    return trace
            raise IndexError(f"{self} doesn't contain trace \"{trace_ref}\"\n"
                             f"Valid traces are {[trc.name for trc in self._traces]}")
        else:
            return self._traces[trace_ref]

    def __getitem__(self, item):
        """Helper function to access traces by using the [ ] operator."""
        return self.get_trace(item)


def tobytes_for_trace(trace: Trace):
    def tobytes(value):
        return value.tobytes()
    return tobytes



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        qsch_editor.py
# Purpose:     Class made to update directly the QSPICE Schematic files
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import math
from pathlib import Path
from typing import Union, Tuple, List
import re
import logging
from spicelib.editor.base_editor import (
    BaseEditor, format_eng, ComponentNotFoundError, ParameterNotFoundError,
    PARAM_REGEX, UNIQUE_SIMULATION_DOT_INSTRUCTIONS
)
__all__ = ('QschEditor', )

from ..utils.detect_encoding import detect_encoding

_logger = logging.getLogger("qspice.QschEditor")

QSCH_HEADER = (255, 216, 255, 219)
QSCH_TEXT_POS = 1
QSCH_TEXT_STR_ATTR = 8
QSCH_COMPONENT_POS = 1
QSCH_SYMBOL_TEXT_REFDES = 0
QSCH_SYMBOL_TEXT_VALUE = 1


class QschReadingError(IOError):
    ...


class QschTag:
    def __init__(self, stream, start):
        assert stream[start] == '«'
        self.start = start
        self.items = []
        self.tokens = []
        i = start + 1
        i0 = i
        while i < len(stream):
            if stream[i] == '«':
                child = QschTag(stream, i)
                i = child.stop
                i0 = i + 1
                self.items.append(child)
            elif stream[i] == '»':
                self.stop = i + 1
                if i > i0:
                    self.tokens.append(stream[i0:i])
                break
            elif stream[i] == ' ' or stream[i] == '\n':
                if i > i0:
                    self.tokens.append(stream[i0:i])
                i0 = i + 1
            elif stream[i] == '"':
                # get all characters until the next " sign
                i += 1
                while stream[i] != '"':
                    i += 1
            elif stream[i] == '(':
                # todo: handle nested parenthesis and also [] and {}
                i += 1
                while stream[i] != ')':
                    i += 1
            i += 1
        else:
            raise IOError("Missing » when reading file")

    def __str__(self):
        """Returns only the first line"""
        return ' '.join(self.tokens)

    def out(self, level):
        spaces = '  ' * level
        if len(self.items):
            return (f"{spaces}«{' '.join(self.tokens)}\n"
                    f"{''.join(tag.out(level+1) for tag in self.items)}"
                    f"{spaces}»\n")
        else:
            return f"{'  ' * level}«{' '.join(self.tokens)}»\n"

    @property
    def tag(self) -> str:
        return self.tokens[0]

    def get_items(self, item) -> List['QschTag']:
        answer = [tag for tag in self.items if tag.tag == item]
        return answer

    def get_attr(self, index: int):
        a = self.tokens[index]
        if a.startswith('(') and a.endswith(')'):
            return tuple(int(x) for x in a[1:-1].split(','))
        elif a.startswith('0x'):
            return int(a[2:], 16)
        elif a.startswith('"') and a.endswith('"'):
            return a[1:-1]
        else:
            return int(a)

    def set_attr(self, index: int, value):
        if isinstance(value, int):
            value_str = str(value)
        elif isinstance(value, str):
            if value.startswith('0x'):
                value_str = value
            else:
                value_str = f'"{value}"'
        else:
            raise ValueError("Object not supported in set_attr")
        self.tokens[index] = value_str

    def get_text(self, label) -> str:
        a = self.get_items(label+':')
        if len(a) != 1:
            raise IndexError(f"Label {label}: not found")
        return a[0].tokens[1]

    def get_text_attr(self, index: int) -> str:
        a = self.tokens[index]
        if a.startswith('"') and a.endswith('"'):
            return a[1:-1]
        else:
            return a


class QschEditor(BaseEditor):
    """Class made to update directly the LTspice ASC files"""

    def __init__(self, asc_file: str):
        self._qsch_file_path = Path(asc_file)
        self.schematic = None
        self._symbols = {}
        if not self._qsch_file_path.exists():
            raise FileNotFoundError(f"File {asc_file} not found")
        # read the file into memory
        self.reset_netlist()

    @property
    def circuit_file(self) -> Path:
        return self._qsch_file_path

    def save_as(self, qsch_file: Union[str, Path]) -> None:
        with open(qsch_file, 'w', encoding="cp1252") as qsch_file:
            _logger.info(f"Writing ASC file {qsch_file}")
            for c in QSCH_HEADER:
                qsch_file.write(chr(c))
            qsch_file.write(self.schematic.out(0))
            qsch_file.write('\n')  # Terminates the new line

    def save_netlist(self, run_netlist_file: Union[str, Path]) -> None:
        if isinstance(run_netlist_file, str):
            run_netlist_file = Path(run_netlist_file)

        if self.schematic is None:
            _logger.error("Empty Schematic information")
            return
        if run_netlist_file.suffix == '.qsch':
            self.save_as(run_netlist_file)
        elif run_netlist_file.suffix in ('.net', '.cir'):
            libraries_to_include = []
            with open(run_netlist_file, 'w') as netlist_file:
                _logger.info(f"Writing NET file {run_netlist_file}")
                netlist_file.write(f'* Netlist generated by PyQSPICE from {self._qsch_file_path}\n')
                for item in self.schematic.items:
                    if item.tag == 'component':
                        component_pos = item.get_attr(QSCH_COMPONENT_POS)
                        orientation = item.get_attr(2)
                        disabled = item.get_attr(3)
                        symbol_tags = item.get_items('symbol')
                        if len(symbol_tags) != 1 or disabled == 1:
                            continue
                        symbol_tag = symbol_tags[0]
                        symbol = symbol_tag.get_text_attr(1)
                        typ = symbol_tag.get_text('type')
                        pins = symbol_tag.get_items('pin')
                        texts = symbol_tag.get_items('text')
                        refdes = texts[0].get_text_attr(QSCH_TEXT_STR_ATTR)
                        nets = [self._find_net_at_pin(component_pos, orientation, pin) for pin in pins]
                        if typ == 'R' or typ == 'D' or typ == 'C' or typ == 'L' or typ == 'V' or typ == 'I':
                            value = texts[1].get_text_attr(QSCH_TEXT_STR_ATTR)
                            if len(texts) > 2:
                                for i in range(2, len(texts)):
                                    value += ' ' + texts[i].get_text_attr(QSCH_TEXT_STR_ATTR)
                            if refdes.startswith(symbol):
                                netlist_file.write(f'{refdes} {" ".join(nets)} {value}\n')
                            else:
                                netlist_file.write(f'{symbol}†{refdes} {" ".join(nets)} {value}\n')
                        elif typ == 'QP' or typ == 'QN':
                            model = texts[1].get_text_attr(QSCH_TEXT_STR_ATTR)
                            netlist_file.write(f'{refdes} {" ".join(nets)} 0 {model} {symbol}\n')
                        else:
                            netlist_file.write(f'Not Found:{typ} {refdes} {component_pos}\n')

                        library_tags = symbol_tag.get_items('library')
                        for lib in library_tags:
                            library_name = lib.get_text_attr(2)
                            if library_name not in libraries_to_include:
                                libraries_to_include.append(library_name)
                    elif item.tag == 'text':
                        is_comment = item.get_attr(4) == 1
                        text = item.get_attr(QSCH_TEXT_STR_ATTR).split('\\n')
                        for line in text:
                            if is_comment:
                                netlist_file.write('* ')
                            netlist_file.write(line.strip() + '\n')
                for library in libraries_to_include:
                    netlist_file.write(f'.lib {library}\n')
                netlist_file.write('.end\n')

    def _find_net_at_pin(self, comp_pos, orientation : int, pin: QschTag) -> str:
        """Returns the net at the given position"""
        pin_pos = pin.get_attr(1)
        hyp = (pin_pos[0] ** 2 + pin_pos[1] ** 2) ** 0.5
        if orientation % 2:
            # in 45º rotations the component is 1.414 times larger
            hyp *= 1.414
        if 0 <= orientation <= 7:
            theta = math.atan2(pin_pos[1], pin_pos[0]) + math.radians(orientation * 45)
            x = comp_pos[0] + round(hyp * math.cos(theta), -2)  # round to multiple of 100
            y = comp_pos[1] + round(hyp * math.sin(theta), -2)
        elif 8 <= orientation <= 15:
            # The component is mirrored on the X axis
            theta = math.atan2(pin_pos[1], pin_pos[0]) + math.radians((orientation - 8) * 45)
            x = comp_pos[0] - round(hyp * math.cos(theta), -2)  # round to multiple of 100
            y = comp_pos[1] + round(hyp * math.sin(theta), -2)
        else:
            raise ValueError(f"Invalid orientation: {orientation}")
        for net in self.schematic.get_items('wire'):
            if net.get_attr(1) == (x, y) or net.get_attr(2) == (x, y):
                net_name = net.get_attr(3)  # Found the net
                return '0' if  net_name == 'GND' else net_name
        else:
            return '####'

    def reset_netlist(self):
        """Reads the ASC file and parses it into memory"""
        with open(self._qsch_file_path, 'r', encoding="cp1252") as asc_file:
            _logger.info(f"Reading QSCH file {self._qsch_file_path}")
            stream = asc_file.read()
        self._parse_qsch_stream(stream)

    def _parse_qsch_stream(self, stream):

        self._symbols.clear()
        _logger.debug("Parsing ASC file")
        header = tuple(ord(c) for c in stream[:4])

        if header != QSCH_HEADER:
            raise QschReadingError("Missing header. The QSCH file should start with: " +
                                   f"{' '.join(f'{c:02X}' for c in QSCH_HEADER)}")

        schematic = QschTag(stream, 4)
        self.schematic = schematic

        components = self.schematic.get_items('component')
        for component in components:
            symbol: QschTag = component.get_items('symbol')[0]
            typ = symbol.get_text('type')
            desc = symbol.get_text('description')
            texts = symbol.get_items('text')
            if len(texts) < 2:
                raise RuntimeError(f"Missing texts in component at coordinates {component.get_attr(1)}")
            refdes = texts[QSCH_SYMBOL_TEXT_REFDES].get_attr(QSCH_TEXT_STR_ATTR)
            value = texts[QSCH_SYMBOL_TEXT_VALUE].get_attr(QSCH_TEXT_STR_ATTR)
            self._symbols[refdes] = {
                'type': typ,
                'description': desc,
                'model': value,
                'tag': component
            }

    def get_component_info(self, component) -> dict:
        """Returns the component information as a dictionary"""
        if component not in self._symbols:
            _logger.error(f"Component {component} not found in ASC file")
            raise ComponentNotFoundError(f"Component {component} not found in ASC file")
        return self._symbols[component]

    def _get_text_matching(self, command, search_expression: re.Pattern):
        command_upped = command.upper()
        text_tags = self.schematic.get_items('text')
        for tag in text_tags:
            line = tag.get_attr(QSCH_TEXT_STR_ATTR)
            if line.upper().startswith(command_upped):
                match = search_expression.search(line)
                if match:
                    return tag, match
        else:
            return None, None

    def get_parameter(self, param: str) -> str:
        param_regex = re.compile(PARAM_REGEX % param, re.IGNORECASE)
        tag, match = self._get_text_matching(".PARAM", param_regex)
        if match:
            return match.group('value')
        else:
            raise ParameterNotFoundError(f"Parameter {param} not found in ASC file")

    def set_parameter(self, param: str, value: Union[str, int, float]) -> None:
        param_regex = re.compile(PARAM_REGEX % param, re.IGNORECASE)
        tag, match = self._get_text_matching(".PARAM", param_regex)
        if match:
            _logger.debug(f"Parameter {param} found in ASC file, updating it")
            if isinstance(value, (int, float)):
                value_str = format_eng(value)
            else:
                value_str = value
            text: str = tag.get_attr(QSCH_TEXT_STR_ATTR)
            match = param_regex.search(text)  # repeating the search, so we update the correct start/stop parameter
            start, stop = match.span(param_regex.groupindex['replace'])
            text =  text[:start] + "{}={}".format(param, value_str) + text[stop:]
            tag.set_attr(QSCH_TEXT_STR_ATTR, text)
            _logger.info(f"Parameter {param} updated to {value_str}")
            _logger.debug(f"Text at {tag.get_attr(QSCH_TEXT_POS)} Updated to {text}")
        else:
            # Was not found so we need to add it,
            _logger.debug(f"Parameter {param} not found in ASC file, adding it")
            x, y = self._get_text_space()
            tag = QschTag(f'«text ({x},{y}) 1 0 0 0x1000000 -1 -1 ".param {param}={value}"»', 0)
            self.schematic.items.append(tag)
            _logger.info(f"Parameter {param} added with value {value}")
            _logger.debug(f"Text added to {tag.get_attr(QSCH_TEXT_POS)} Added: {tag.get_attr(QSCH_TEXT_STR_ATTR)}")

    def set_component_value(self, device: str, value: Union[str, int, float]) -> None:
        if isinstance(value, str):
            value_str = value
        else:
            value_str = format_eng(value)
        self.set_element_model(device, value_str)

    def set_element_model(self, device: str, model: str) -> None:
        comp_info = self.get_component_info(device)
        component: QschTag = comp_info['tag']
        symbol: QschTag = component.get_items('symbol')[0]
        texts = symbol.get_items('text')
        assert texts[QSCH_SYMBOL_TEXT_REFDES].get_attr(QSCH_TEXT_STR_ATTR) == device
        texts[QSCH_SYMBOL_TEXT_VALUE].set_attr(QSCH_TEXT_STR_ATTR, model)
        self._symbols[device]['model'] = model
        _logger.info(f"Component {device} updated to {model}")
        _logger.debug(f"Component at :{component.get_attr(1)} Updated")

    def get_component_value(self, element: str) -> str:
        comp_info = self.get_component_info(element)
        if "model" not in comp_info:
            _logger.error(f"Component {element} does not have a Value attribute")
            raise ComponentNotFoundError(f"Component {element} does not have a Value attribute")
        return comp_info["model"]

    def get_components(self, prefixes='*') -> list:
        if prefixes == '*':
            return list(self._symbols.keys())
        return [k for k in self._symbols.keys() if k[0] in prefixes]

    def remove_component(self, designator: str):
        comp_info = self.get_component_info(designator)
        component: QschTag = comp_info['tag']
        self.schematic.items.remove(component)

    def _get_text_space(self):
        """
        Returns the coordinate on the Schematic File canvas where a text can be appended.
        """
        first = True
        for tag in self.schematic.items:
            if tag.tag in ('component', 'net', 'text'):
                x1, y1 = tag.get_attr(1)
                x2, y2 = x1, y1  # todo: the whole component primitives
            elif tag.tag == 'wire':
                x1, y1 = tag.get_attr(1)
                x2, y2 = tag.get_attr(2)
            else:
                continue  # this avoids executing the code below when no coordinates are found
            if first:
                min_x = min(x1, x2)
                max_x = max(x1, x2)
                min_y = min(y1, y2)
                max_y = max(y1, y2)
                first = False
            else:
                min_x = min(min_x, x1, x2)
                max_x = max(max_x, x1, x2)
                min_y = min(min_y, y1, y2)
                max_y = max(max_y, y1, y2)

        if first:
            return 0, 0  # If no coordinates are found, we return the origin
        else:
            return min_x, min_y - 240  # Setting the text in the bottom left corner of the canvas

    def add_instruction(self, instruction: str) -> None:
        instruction = instruction.strip()  # Clean any end of line terminators
        command = instruction.split()[0].upper()

        if command in UNIQUE_SIMULATION_DOT_INSTRUCTIONS:
            # Before adding new instruction, if it is a unique instruction, we just replace it
            for text_tag in self.schematic.get_items('text'):
                text = text_tag.get_attr(QSCH_TEXT_STR_ATTR)
                command = text.split()[0].upper()
                if command in UNIQUE_SIMULATION_DOT_INSTRUCTIONS:
                    text_tag.set_attr(QSCH_TEXT_STR_ATTR, instruction)
                    return  # Job done, can exit this method

        elif command.startswith('.PARAM'):
            raise RuntimeError('The .PARAM instruction should be added using the "set_parameter" method')
        # If we get here, then the instruction was not found, so we need to add it
        x, y = self._get_text_space()
        tag = QschTag(f'«text ({x},{y}) 1 0 0 0x1000000 -1 -1 "{instruction}"»', 0)
        self.schematic.items.append(tag)

    def remove_instruction(self, instruction: str) -> None:
        for text_tag in self.schematic.get_items('text'):
            text = text_tag.get_attr(QSCH_TEXT_STR_ATTR)
            if instruction in text:
                self.schematic.items.remove(text_tag)
                _logger.info(f'Instruction "{instruction}" removed')
                return  # Job done, can exit this method

        msg = f'Instruction "{instruction}" not found'
        _logger.error(msg)

    def remove_Xinstruction(self, search_pattern: str) -> None:
        regex = re.compile(search_pattern, re.IGNORECASE)
        instr_removed = False
        for text_tag in self.schematic.get_items('text'):
            text = text_tag.get_attr(QSCH_TEXT_STR_ATTR)
            if regex.match(text):
                self.schematic.items.remove(text_tag)
                _logger.info(f'Instruction "{text}" removed')
                instr_removed = True
        if not instr_removed:
            msg = f'Instruction matching "{search_pattern}" not found'
            _logger.error(msg)



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sim_client.py
# Purpose:     Example of a client to the SimServer
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-02-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import os.path
import zipfile
import logging

# In order for this, to work, you need to have a server running. To start a server, run the following command:
# python -m spicelib.run_server --port 9000 --parallel 4 --output ./temp

_logger = logging.getLogger("spicelib.SimClient")
_logger.setLevel(logging.DEBUG)

from spicelib.client_server.sim_client import SimClient

server = SimClient('http://localhost', 9000)
print(server.session_id)
runid = server.run("./testfiles/testfile.net")
print("Got Job id", runid)
for runid in server:  # Ma
    zip_filename = server.get_runno_data(runid)
    print(f"Received {zip_filename} from runid {runid}")
    with zipfile.ZipFile(zip_filename, 'r') as zipf:  # Extract the contents of the zip file
        print(zipf.namelist())  # Debug printing the contents of the zip file
        zipf.extract(zipf.namelist()[0])  # Normally the raw file comes first
    os.remove(zip_filename)  # Remove the zip file

server.close_session()
print("Finished")
# server.server.stop_server()  # This will terminate the server



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sim_client.py
# Purpose:     Tool used to launch a Spice simulation in batch mode.
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-02-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import os.path
import zipfile
import xmlrpc.client
import io
import pathlib
import time
from collections import OrderedDict
from dataclasses import dataclass
import logging
from typing import Union

_logger = logging.getLogger("spicelib.SimClient")


class SimClientInvalidRunId(LookupError):
    """Raised when asking for a run_no that doesn't exist"""
    ...

@dataclass
class JobInformation:
    """Contains information about pending simulation jobs"""
    run_number: int  # The run id that is returned by the Server and which identifies the server
    file_dir: pathlib.Path

# class RunIterator(object):
#
#     def __init__(self, client, timeout):
#         self.client = client
#         self.timeout = timeout
#
#     def __iter__(self):
#         return self
#
#     def __next__(self):
#         return self.client.__next__()


class SimClient(object):
    """
    Class used for launching simulations in a Spice Simulation Server.
    A Spice Simulation Server is a machine running a script with an active SimServer object.

    This class only implement basic level handshaking with a single simulation Server.
    Upon instance, it will establish a connection with Simulation Server. This connection is kept
    alive during the whole live of this object.

    The run() method will transfer the netlist for the server, execute a simulation and transfer the simulation results
    back to the client.

    Data is returned from the server inside a zipfie which is copied into the directory defined when the job was
    created, /i.e./ run() method called.

    Two lists are kept by this class:

        * A list of started jobs (started_jobs) and,

        * a list with finished jobs on the server, but, which haven't been yet transferred to the client (stored_jobs).

    This distinction is important because the data is erased on the server side when the data is transferred.

    This class implements an iterator that is to be used for retrieving the job. See the example below.
    The iterator polls the server with a time interval defined by the attribute ``minimum_time_between_server_calls``.
    This attribute is set to 0.2 seconds by default, but it can be overriden.

    Usage:

    .. code-block:: python

        import zipfile
        from PySpice.sim.sim_client import SimClient

        server = SimClient('http://localhost', 9000)  # Use another computer address.
        print(server.session_id)
        runid = server.run("../../tests/testfile.net")
        print("Got Job id", runid)

        for runid in server:   # may not arrive in the same order as runids were launched
            zip_filename = server.get_runno_data(runid)
            print(f"Received {zip_filename} from runid {runid}")

            with zipfile.ZipFile(zip_filename, 'r') as zipf:  # Extract the contents of the zip file
                print(zipf.namelist())  # Debug printing the contents of the zip file
                zipf.extract(zipf.namelist()[0])  # Normally the raw file comes first

    NOTE: More elaborate algorithms such as managing multiple servers will be done on another class.
    """

    def __init__(self, host_address, port):
        self.server = xmlrpc.client.ServerProxy(f'{host_address}:{port}')
        self.session_id = self.server.start_session()
        _logger.info(f"Started {self.session_id}")
        self.started_jobs = OrderedDict()  # This list keeps track of started jobs on the server
        self.stored_jobs = OrderedDict()  # This list keeps track of finished simulations that haven't yet been transferred.
        self.completed_jobs = 0
        self.minimum_time_between_server_calls = 0.2  # Minimum time between server calls
        self._last_server_call = time.time()

    def __del__(self):
        self.close_session()

    def run(self, circuit):
        """
        Sends the netlist identified with the argument "circuit" to the server, and it receives a run identifier
        (runno). Since the server can receive requests from different machines, this identifier is not guaranteed to be
        sequential.

        :param circuit: path to the netlist file containing the simulation directives.
        :type circuit: pathlib.Path or str
        :returns: identifier on the server of the simulation.
        :rtype: int
        """
        circuit_path = pathlib.Path(circuit)
        circuit_name = circuit_path.name
        if os.path.exists(circuit):
            # Create a buffer to store the zip file in memory
            zip_buffer = io.BytesIO()

            # Create the zip file in memory
            with zipfile.ZipFile(zip_buffer, 'w', zipfile.ZIP_DEFLATED) as zip_file:
                zip_file.write(circuit, circuit_name)  # Makes sure it writes it to the root of the zipfile

            # Reset the buffer position to the start
            zip_buffer.seek(0)

            # Read the zip file from the buffer and send it to the server
            zip_data = zip_buffer.read()

            run_id = self.server.run(self.session_id, circuit_name, zip_data)
            job_info = JobInformation(run_number=run_id, file_dir=circuit_path.parent)
            self.started_jobs[run_id] = job_info
            return run_id
        else:
            _logger.error(f"Circuit {circuit} doesn't exit")
            return -1
        
    def get_runno_data(self, runno) -> Union[str, None]:
        """
        Returns the simulation output data inside a zip file name.

        :rtype: str
        """
        if runno not in self.stored_jobs:
            raise SimClientInvalidRunId(f"Invalid Job id {runno}")

        zip_filename, zipdata = self.server.get_files(self.session_id, runno)
        job = self.stored_jobs.pop(runno)  # Removes it from stored jobs
        self.completed_jobs += 1
        if zip_filename != '':
            store_path = job.file_dir / zip_filename
            with open(store_path, 'wb') as f:
                f.write(zipdata.data)
            return store_path
        else:
            return None

    def __iter__(self):
        return self
    
    def __next__(self):
        while len(self.started_jobs) > 0:
            status = self.server.status(self.session_id)
            if len(status) > 0:
                runno = status.pop(0)
                self.stored_jobs[runno] = self.started_jobs.pop(runno)  # Job is taken out of the started jobs list and
                # is added to the stored jobs
                return runno
            else:
                now = time.time()
                delta = self.minimum_time_between_server_calls - (now - self._last_server_call)
                if delta > 0:
                    time.sleep(delta)  # Go asleep for a sec
                self._last_server_call = now

        # when there are no pending jobs left, exit the iterator    
        raise StopIteration

    def close_session(self):
        _logger.info(f"closing session {self.session_id}")
        self.server.close_session(self.session_id)



from spicelib.log.ltsteps import LTSpiceLogReader


log = LTSpiceLogReader("2_NJF_Bias_dB.log")
print(log.fourier)
print(log.stepset)
print(log.dataset)
print(log.step_count)
log.export_data("2_NJF_Bias_dB.tlog")


from spicelib import AscEditor, SimRunner  # Imports the class that manipulates the asc file
from spicelib.sim.tookit.quick_sensitivity_analysis import QuickSensitivityAnalysis
from spicelib.simulators.ltspice_simulator import LTspice

sallenkey = AscEditor("./testfiles/sallenkey.asc")  # Reads the asc file into memory
runner = SimRunner(simulator=LTspice, output_folder='./temp_sens')  # Instantiates the runner with a temp folder set
sa = QuickSensitivityAnalysis(sallenkey, runner)  # Instantiates the Worst Case Analysis class

# The following lines set the default tolerances for the components
sa.set_tolerance('R', 0.01)  # 1% tolerance
sa.set_tolerance('C', 0.1)  # 10% tolerance
sa.set_tolerance('V', 0.1)  # 10% tolerance. For Worst Case analysis, the distribution is irrelevant

# Some components can have a different tolerance
sa.set_tolerance('R1', 0.05)  # 5% tolerance for R1 only. This only overrides the default tolerance for R1

# Tolerances can be set for parameters as well.
sa.set_parameter_deviation('Vos', 3e-4, 5e-3)

# Finally the netlist is saved to a file
sa.save_netlist('./testfiles/sallenkey_sa.asc')


sa.run_testbench()  # Runs the simulation with splits of 100 runs each
logs = sa.read_logfiles()   # Reads the log files and stores the results in the results attribute
logs.export_data('./temp_sens/data.csv')  # Exports the data to a csv file

print("Sensitivity results:")
sens = sa.get_sensitivity_data('*', 'fcut')
for comp, value in sens.items():
    print(f"{comp}: {value:.2f}%")

sa.cleanup_files()  # Deletes the temporary files



# coding=utf-8

import sys
sys.path.insert(0, '..')    # This is to allow the import from the spicelib folder
from spicelib import SimRunner, SpiceEditor
from spicelib.sim.process_callback import ProcessCallback  # Importing the ProcessCallback class type


class CallbackProc(ProcessCallback):
    """Class encapsulating the callback function. It can have whatever name."""

    @staticmethod
    def callback(raw_file, log_file):
        print("Handling the simulation data of ""%s"", log file ""%s""" % (raw_file, log_file))
        # Doing some processing here
        return "Parsed Result of ""%s""" % raw_file + ", log file ""%s""" % log_file


if __name__ == "__main__":
    from spicelib.simulators.ltspice_simulator import LTspice
    runner = SimRunner(output_folder='./temp_batch4', simulator=LTspice)  # Configures the simulator to use and output
    # folder

    netlist = SpiceEditor("./testfiles/Batch_Test.net")  # Open the Spice Model, and creates the .net
    # set default arguments
    netlist.set_parameters(res=0, cap=100e-6)
    netlist.set_component_value('R2', '2k')  # Modifying the value of a resistor
    netlist.set_component_value('R1', '4k')
    netlist.set_element_model('V3', "SINE(0 1 3k 0 0 0)")  # Modifying the
    netlist.set_component_value('XU1:C2', 20e-12)  # modifying a
    # define simulation
    netlist.add_instructions(
        "; Simulation settings",
        ";.param run = 0"
    )
    netlist.set_parameter('run', 0)

    for opamp in ('AD712', 'AD820'):
        netlist.set_element_model('XU1', opamp)
        for supply_voltage in (5, 10, 15):
            netlist.set_component_value('V1', supply_voltage)
            netlist.set_component_value('V2', -supply_voltage)
            # overriding the automatic netlist naming
            run_netlist_file = "{}_{}_{}.net".format(netlist.netlist_file.stem, opamp, supply_voltage)
            runner.run(netlist, run_filename=run_netlist_file, callback=CallbackProc)

    for result in runner:
        print(result)  # Prints the result of the callback function

    netlist.reset_netlist()
    netlist.add_instructions(   # Adding additional instructions
        "; Simulation settings",
        ".ac dec 30 10 1Meg",
        ".meas AC Gain MAX mag(V(out)) ; find the peak response and call it ""Gain""",
        ".meas AC Fcut TRIG mag(V(out))=Gain/sqrt(2) FALL=last"
    )

    raw, log = runner.run_now(netlist, run_filename="no_callback.net")
    CallbackProc.callback(raw, log)

    results = runner.wait_completion(1, abort_all_on_timeout=True)

    # Sim Statistics
    print('Successful/Total Simulations: ' + str(runner.okSim) + '/' + str(runner.runno))





#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        qspice_simulator.py
# Purpose:     Represents QSPICE
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     26-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import sys
import os

from pathlib import Path
import logging
_logger = logging.getLogger("spicelib.QSpiceSimulator")

from ..sim.simulator import Simulator, run_function, SpiceSimulatorError


class Qspice(Simulator):
    """Stores the simulator location and command line options and is responsible for generating netlists and running
    simulations."""
    raw_extension = '.qraw'  # In QSPICE all traces have double precision. This means that qraw files are not compatible
    # with LTSPICE

    """Searches on the any usual locations for a simulator"""
    if sys.platform == "linux":
        spice_exe = []
        _logger.warning("QSPICE is not available for Linux")
    elif sys.platform == "darwin":
        spice_exe = []
        _logger.warning("QSPICE is not available for MacOS")
    else:  # Windows
        for exe in (  # Placed in order of preference. The first to be found will be used.
                os.path.expanduser(r"~\AppData\Local\Programs\Qspice\QSPICE64.exe"),
                r"C:\Program Files\QSPICE\QSPICE64.exe",
        ):
            if os.path.exists(exe):
                _logger.debug(f"Using Qspice installed in : '{exe}' ")
                spice_exe = [exe]
                break
        else:
            spice_exe = []
            _logger.warning("QSPICE not found in the usual locations. Please install it and try again.")

        process_name = "QSPICE64.exe"

    qspice_args = {
        'ASCII'     : ['-ASCII'],  # Use ASCII file format for the output data(.qraw) file.
        'binary'    : ['-binary'],  # Use binary file format for the output data(.qraw) file.
        'BSIM1'    : ['-BSIM1'],  # Use the charge-conserving BSIM1 charge model for MOS1, MOS2, and MOS3.
        'Meyer'    : ['-Meyer'],  # Use the Meyer Capacitance model for MOS1, MOS2, and MOS3.
        'o'         : ['-o', '<path>'],  # Specify the name of a file for the console output.
        # 'p'         : ['-p'],  # Take the netlist piped from stdin. Not used in this implementation.
        'ProtectSelections': ['-ProtectSelections', '<path>'],  # Protect sections marked with .prot/.unprot with encryption.
        'ProtectSubcircuits': ['-ProtectSubcircuits', '<path>'],  # Protect the body of subcircuits with encryption.
        'r'       : ['-r', '<path>'],  # Specify the name of the output data(.qraw) file.
    }

    @classmethod
    def valid_switch(cls, switch, path='') -> list:
        """
        Validates a command line switch. The following options are available for QSPICE:
            * 'ASCII': Use ASCII file format for the output data(.qraw) file.

            * 'binary': Use binary file format for the output data(.qraw) file.

            * 'BSIM1': Use the charge-conserving BSIM1 charge model for MOS1, MOS2, and MOS3.

            * 'Meyer': Use the Meyer Capacitance model for MOS1, MOS2, and MOS3.

            * 'o <path>': Specify the name of a file for the console output.

            * 'ProtectSelections <path>': Protect sections marked with .prot/.unprot with encryption.

            * 'ProtectSubcircuits <path>': Protect the body of subcircuits with encryption.

            * 'r <path>': Specify the name of the output data(.qraw) file.


        :param switch: switch to be added. If the switch is not on the list above, it should be correctly formatted with
                       the preceding '-' switch
        :type switch: str
        :param path: path to the file related to the switch being given.
        :type path: str, optional
        :return: Nothing
        :rtype: None
        """
        if switch in cls.qspice_args:
            switches = cls.qspice_args[switch]
            switches = [switch.replace('<path>', path) for switch in switches]
            return switches
        else:
            raise ValueError("Invalid switch for class ")

    @classmethod
    def run(cls, netlist_file, cmd_line_switches, timeout):
        if not cls.spice_exe:
            _logger.error("================== ALERT! ====================")
            _logger.error("Unable to find the QSPICE executable.")
            _logger.error("A specific location of the QSPICE can be set")
            _logger.error("using the create_from(<location>) class method")
            _logger.error("==============================================")
            raise SpiceSimulatorError("Simulator executable not found.")
        log_file = Path(netlist_file).with_suffix('.log').as_posix()
        cmd_run = cls.spice_exe + ['-o', log_file] + [netlist_file] + cmd_line_switches
        # start execution
        return run_function(cmd_run, timeout=timeout)



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        run_server.py
# Purpose:     A Command Line Interface to run the LTSpice Server
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     10-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import sys
import argparse
from spicelib.client_server.sim_server import SimServer
import time
import keyboard


def main():
    parser = argparse.ArgumentParser(
        description="Run the LTSpice Server. This is a command line interface to the SimServer class."
                    "The SimServer class is used to run simulations in parallel using a server-client architecture."
                    "The server is a machine that runs the SimServer class and the client is a machine that runs the "
                    "SimClient class."
                    "The argument is the simulator to be used (LTSpice, NGSpice, XYCE, etc.)"
    )
    parser.add_argument('simulator', type=str, default="LTSpice",
                        help="Simulator to be used (LTSpice, NGSpice, XYCE, etc.)")
    parser.add_argument("-p", "--port", type=int, default=9000,
                        help="Port to run the server. Default is 9000")
    parser.add_argument("-o", "--output", type=str, default=".",
                        help="Output folder for the results. Default is the current folder")
    parser.add_argument("-l", "--parallel", type=int, default=4,
                        help="Maximum number of parallel simulations. Default is 4")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    if args.parallel < 1:
        args.parallel = 1

    if args.simulator == "LTSpice":
        from spicelib.simulators.ltspice_simulator import LTspice
        simulator = LTspice
    elif args.simulator == "NGSpice":
        from spicelib.simulators.ngspice_simulator import NGspiceSimulator
        simulator = NGspiceSimulator
    elif args.simulator == "XYCE":
        from spicelib.simulators.xyce_simulator import XyceSimulator
        simulator = XyceSimulator
    else:
        raise ValueError(f"Simulator {args.simulator} is not supported")
        exit(-1)

    print("Starting Server")
    server = SimServer(simulator, parallel_sims=args.parallel, output_folder=args.output, port=args.port)
    print("Server Started. Press and hold 'q' to stop")
    while server.running():
        time.sleep(0.2)
        # Check whether a key was pressed
        if keyboard.is_pressed('q'):
            server.stop_server()
            break


if __name__ == "__main__":
    import logging
    log1 = logging.getLogger("spicelib.ServerSimRunner")
    log2 = logging.getLogger("spicelib.SimServer")
    log3 = logging.getLogger("spicelib.SimRunner")
    log4 = logging.getLogger("spicelib.RunTask")
    log1.setLevel(logging.INFO)
    log2.setLevel(logging.INFO)
    log3.setLevel(logging.INFO)
    log4.setLevel(logging.INFO)
    main()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        rawplot.py
# Purpose:     Make a plot of the data in the raw file
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     02-09-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------


def main():
    """Uses matplotlib to plot the data in the raw file"""
    import sys
    import matplotlib
    import matplotlib.pyplot as plt

    from numpy import abs as mag, angle as phase_np
    from numpy import angle, arange

    from spicelib import RawRead

    def what_to_units(whattype):
        """Determines the unit to display on the plot Y axis"""
        if 'voltage' in whattype:
            return 'V'
        if 'current' in whattype:
            return 'A'

    matplotlib.use('tkagg')

    if len(sys.argv) > 2:
        raw_filename = sys.argv[1]
        trace_names = sys.argv[2:]
    elif len(sys.argv) > 1:
        raw_filename = sys.argv[1]
        trace_names = '*'  # All traces
    else:
        print("Usage: rawplot.py RAW_FILE TRACE_NAME")
        print("TRACE_NAME is the traces to plot")
        sys.exit(-1)

    LTR = RawRead(raw_filename, trace_names, verbose=True)
    for param, value in LTR.raw_params.items():
        print("{}: {}{}".format(param, " " * (20 - len(param)), str(value).strip()))

    if trace_names == '*':

        print("Reading all the traces in the raw file")
        trace_names = LTR.get_trace_names()

    traces = [LTR.get_trace(trace) for trace in trace_names]
    if LTR.axis is not None:
        steps_data = LTR.get_steps()
    else:
        steps_data = [0]
    print("Steps read are :", list(steps_data))

    n_axis = len(traces)

    fig, axis_set = plt.subplots(nrows=n_axis, ncols=1, sharex='all')

    if n_axis == 1:
        axis_set = [axis_set]  # Needs to have a list of axis

    for i, trace in enumerate(traces):
        ax = axis_set[i]

        ax.grid(True)
        if 'log' in LTR.flags:
            ax.set_xscale('log')
        for step_i in steps_data:
            if LTR.axis:
                x = LTR.get_axis(step_i)
            else:
                x = arange(LTR.nPoints)
            y = traces[i].get_wave(step_i)
            label = f"{trace.name}:{steps_data[step_i]})"
            if 'complex' in LTR.flags:
                x = mag(x)
                ax.set_yscale('log')
                y = mag(y)
                ax.yaxis.label.set_color('blue')
                ax.set(ylabel=label+'(dB)')
                ax.plot(x, y)
                ax_phase = ax.twinx()
                y = phase_np(y, deg=True)
                ax_phase.plot(x, y, color='red', linestyle='-.')
                ax_phase.yaxis.label.set_color('red')
                ax_phase.set(ylabel=label+'Phase (o)')
                # title = f"{trace.name} Phase [deg]"
                # ax.set_title(title)
            else:
                ax.plot(x, y)
                ax.set(ylabel=label)
                # title = f"{trace.name} [{what_to_units(trace.whattype)}]"
                # ax.set_title(title)
    fig.tight_layout()
    plt.show()
'''
'''
# out = open("RAW_TEST_out_test1.txt", 'w')
#
# for step in LTR.get_steps():
#     for x in range(len(LTR[0].data)):
#         out.write("%s, %e, %e\n" % (step, LTR[0].data[x], LTR[2].data[x]))
# out.close()
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        Histogram.py
# Purpose:     Make an histogram plot based on data provided by the user
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     17-01-2017
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
This module uses matplotlib to plot a histogram of a gaussian distribution and calculates the project n-sigma interval.

The data can either be retrieved from the clipboard or from a text file. Use the following command line text to call
this module.

.. code-block:: text

    python -m spicelib.Histogram [options] [data_file] TRACE

The help can be obtained by calling the script without arguments

.. code-block:: text

    Usage: Histogram.py [options] LOG_FILE TRACE

    Options:
      --version             show program's version number and exit
      -h, --help            show this help message and exit
      -s SIGMA, --sigma=SIGMA
                            Sigma to be used in the distribution fit. Default=3
      -n NBINS, --nbins=NBINS
                            Number of bins to be used in the histogram. Default=20
      -c FILTERS, --condition=FILTERS
                            Filter condition writen in python. More than one
                            expression can be added but each expression should be
                            preceded by -c. EXAMPLE: -c V(N001)>4 -c parameter==1
                            -c  I(V1)<0.5
      -f FORMAT, --format=FORMAT
                            Format string for the X axis. Example: -f %3.4f
      -t TITLE, --title=TITLE
                            Title to appear on the top of the histogram.
      -r RANGE, --range=RANGE
                            Range of the X axis to use for the histogram in the
                            form min:max. Example: -r -1:1
      -C, --clipboard       If the data from the clipboard is to be used.
      -i IMAGEFILE, --image=IMAGEFILE
                            Name of the image File. extension 'png'


"""
__author__ = "Nuno Canto Brum <me@nunobrum.com>"
__copyright__ = "Copyright 2017, Fribourg Switzerland"

from spicelib.utils.detect_encoding import EncodingDetectError, detect_encoding


def main():
    import numpy as np
    from scipy.stats import norm
    import matplotlib.pyplot as plt

    from optparse import OptionParser

    usage = "usage: %prog [options] LOG_FILE TRACE"
    opts = OptionParser(usage=usage, version="%prog 0.1")
    # opts.add_option('v', "var", action="store", type="string", dest="trace", help="The trace to be used in the histogram")
    opts.add_option('-s', "--sigma", action="store", type="int", dest="sigma", default=3,
                    help="Sigma to be used in the distribution fit. Default=3")
    opts.add_option('-n', "--nbins", action="store",  type="int", dest="nbins", default=20,
                    help="Number of bins to be used in the histogram. Default=20")
    opts.add_option('-c', "--condition", action="append", type="string", dest="filters",
                    help="Filter condition writen in python. More than one expression can be added but each expression "
                         "should be preceded by -c.\n" +
                         "EXAMPLE: -c V(N001)>4 -c parameter==1 -c I(V1)<0.5"
                          "Note: whe parsing log files, the > and < operators are not supported."
                    )
    opts.add_option('-f', "--format", action="store", type="string", dest="format",
                    help="Format string for the X axis. Example: -f %3.4f")
    # opts.add_option('-p', "--scaling",action="store", type="string", dest="prescaling",
    # help="Prescaling function to be applied to the input value.")
    opts.add_option('-t', "--title", action="store", type="string", dest="title",
                    help="Title to appear on the top of the histogram.")
    opts.add_option('-r', "--range", action="store", type="string", dest="range",
                    help="Range of the X axis to use for the histogram in the form min:max. Example: -r -1:1")
    opts.add_option('-C', "--clipboard", action="store_true", dest="clipboard",
                    help="If the data from the clipboard is to be used.")
    # opts.add_option('-x', "--xname", action="store", dest="xname", help="Name for the variable displayed")
    opts.add_option('-o', "--output", action="store", type="string", dest="imagefile",
                    help="Output the image to a file. Argument is the name of the image File with png extension.\n"
                         "Example: -o image.png")
    opts.add_option('-1', "--nonorm", action="store_false", dest="normalized", default=True,
                    help="Doesn't normalize the histogram so that area of the bell curve is 1.")
    (options, args) = opts.parse_args()

    values = []


    if options.clipboard:
        try:
            import clipboard
        except ImportError:
            print("Failed to load clipboard package. Use PiP to install it.")
            exit(1)
        if len(args) > 0:
            TRACE = args[-1]
        else:
            TRACE = "var"
        text = clipboard.paste()
        for line in text.split('\n'):
            try:
                values.append(float(line))
            except ValueError:
                print("Failed to convert line: '", line, "'")
    elif len(args) == 0:
        opts.print_help()
        exit(-1)
    else:
        if len(args) < 2:
            opts.error("Wrong number of parameters. Need to give the filename and the trace.")
            opts.print_help()
            exit(-1)
        TRACE = args[1]
        logfile = args[0]

        if not options.filters is None:
            print("Filters Applied:", options.filters)
        else:
            print("No filters defined")

        if logfile.endswith(".log"):
            # Maybe it is a LTSpice log file
            from spicelib.log.ltsteps import LTSpiceLogReader
            try:
                log = LTSpiceLogReader(logfile)
            except EncodingDetectError:
                print("Failed to load file '%s'. Use LTSteps first to convert to tlog format" % logfile)
                exit(-1)
            else:
                if options.filters is None:
                    values = log.get_measure_values_at_steps(TRACE, None)
                else:
                    # This implementation only allows equal operators
                    filters = {}
                    for expression in options.filters:
                        lhs_rhs = expression.split("==")
                        if len(lhs_rhs) == 2:
                            filters[lhs_rhs[0]] = float(lhs_rhs[1])
                        else:
                            print("Unsupported comparison operator in reading .log files.")
                            print("For enhanced comparators convert the file to tlog using LTsteps script")
                    log.steps_with_conditions(**filters)
                    values = log.get_measure_values_at_steps(TRACE, options.filters)

        if len(values) == 0:
            encoding = detect_encoding(logfile)
            print("Loading file '%s' with encoding '%s'" % (logfile, encoding))
            log = open(logfile, 'r', encoding=encoding)
            header = log.readline().rstrip('\r\n')
            vars = header.split('\t')
            try:
                sav_col = vars.index(TRACE)
            except ValueError:
                log.close()
                print("File '%s' doesn't have trace '%s'" % (logfile, TRACE))
                print("LOG FILE contains %s" % vars)
                exit(-1)

            if (options.filters is None) or (len(options.filters) == 0):
                for line in log:
                    vs = line.split('\t')
                    values.append(float(vs[sav_col]))
            else:
                for line in log:
                    vs = map(float, line.split('\t'))
                    env = dict(zip(vars,vs))

                    for expression in options.filters:
                        test = eval(expression, None, env)
                        if test == False:
                            break
                    else:
                        values.append(float(env[TRACE]))
            log.close()

    if len(values) == 0:
        print("No elements found")
    elif len(values) < options.nbins:
        print("Not enough elements for an histogram."
              f"Only found {len(values)} elements. Histogram is specified for {options.nbins} bins")
    else:
        x = np.array(values, dtype=float)
        mu = x.mean()
        mn = x.min()
        mx = x.max()
        sd = np.std(x)
        sigmin = mu - options.sigma*sd
        sigmax = mu + options.sigma*sd

        if options.range is None:
            # Automatic calculation of the range
            axisXmin = mu - (options.sigma + 1) * sd
            axisXmax = mu + (options.sigma + 1) * sd

            if mn < axisXmin:
                axisXmin = mn

            if mx > axisXmax:
                axisXmax = mx
        else:
            try:
                smin, smax = options.range.split(":")
                axisXmin = float(smin)
                axisXmax = float(smax)
            except:
                opts.error("Invalid range setting")
                exit(-1)
        if options.format:
            fmt = options.format
        else:
            fmt = "%f"

        print("Collected %d elements" % len(values))
        print("Distributing in %d bins" % options.nbins)
        print("Minimum is " + fmt % mn)
        print("Maximum is " + fmt % mx)
        print("Mean is " + fmt % mu)
        print("Standard Deviation is " + fmt % sd)
        print(("Sigma %d boundaries are " + fmt + " and " + fmt) % (options.sigma, sigmin, sigmax))
        n, bins, patches = plt.hist(x, options.nbins, density=options.normalized, facecolor='green', alpha=0.75,
                                    range=(axisXmin, axisXmax))
        axisYmax = n.max() * 1.1

        if options.normalized:
            # add a 'best fit' line
            y = norm.pdf(bins, mu, sd)
            l = plt.plot(bins, y, 'r--', linewidth=1)
            plt.axvspan(mu - options.sigma*sd, mu + options.sigma*sd, alpha=0.2, color="cyan")
            plt.ylabel('Distribution [Normalised]')
        else:
            plt.ylabel('Distribution')
        plt.xlabel(TRACE)

        if options.title is None:
            title = (r'$\mathrm{Histogram\ of\ %s:}\ \mu='+fmt+r',\ stdev='+fmt+r',\ \sigma=%d$') % (TRACE, mu, sd, options.sigma)
        else:
            title = options.title
        plt.title(title)

        plt.axis([axisXmin, axisXmax, 0, axisYmax ])
        plt.grid(True)
        if options.imagefile is not None:
            plt.savefig(options.imagefile)
        else:
            plt.show()
main()



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        asc_editor.py
# Purpose:     Class made to update directly the LTspice ASC files
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
from pathlib import Path
from typing import Union, Tuple, List
import re
import logging
from .base_editor import BaseEditor, format_eng, ComponentNotFoundError, ParameterNotFoundError, PARAM_REGEX, \
    UNIQUE_SIMULATION_DOT_INSTRUCTIONS

_logger = logging.getLogger("spicelib.AscEditor")

TEXT_REGEX = re.compile(r"TEXT (-?\d+)\s+(-?\d+)\s+(Left|Right|Top|Bottom)\s\d+\s*(?P<type>[!;])(?P<text>.*)",
                        re.IGNORECASE)
TEXT_REGEX_X = 1
TEXT_REGEX_Y = 2
TEXT_REGEX_ALIGN = 3
TEXT_REGEX_TYPE = 4
TEXT_REGEX_TEXT = 5

END_LINE_TERM = "\n"


class AscEditor(BaseEditor):
    """Class made to update directly the LTspice ASC files"""

    def __init__(self, asc_file: str):
        self._asc_file_path = Path(asc_file)
        self._asc_file_lines: List[str] = []
        self._symbols = {}
        self._texts = []  # This is only here to avoid cycling over the netlist everytime we need to retrieve the texts
        if not self._asc_file_path.exists():
            raise FileNotFoundError(f"File {asc_file} not found")
        # read the file into memory
        self.reset_netlist()

    @property
    def circuit_file(self) -> Path:
        return self._asc_file_path

    def save_netlist(self, run_netlist_file: Union[str, Path]) -> None:
        if isinstance(run_netlist_file, str):
            run_netlist_file = Path(run_netlist_file)
        run_netlist_file = run_netlist_file.with_suffix(".asc")
        with open(run_netlist_file, 'w') as asc_file:
            _logger.info(f"Writing ASC file {run_netlist_file}")
            asc_file.writelines(self._asc_file_lines)

    def reset_netlist(self):
        with open(self._asc_file_path, 'r') as asc_file:
            _logger.info(f"Reading ASC file {self._asc_file_path}")
            self._asc_file_lines = asc_file.readlines()
        self._parse_asc_file()

    def _parse_asc_file(self):
        symbol_attributes = {}
        self._symbols.clear()
        self._texts.clear()
        _logger.debug("Parsing ASC file")
        for line_no, line in enumerate(self._asc_file_lines):
            if line.startswith("SYMBOL"):
                tokens = line.split()
                if symbol_attributes:
                    self._symbols[symbol_attributes["InstName"]] = symbol_attributes
                symbol_attributes = {'Name': tokens[1], 'line': line_no}
            elif line.startswith("SYMATTR"):
                tokens = line.split()
                if symbol_attributes:
                    symbol_attributes[tokens[1]] = tokens[2]
            elif line.startswith("TEXT"):
                match = TEXT_REGEX.match(line)
                if match and match.group(TEXT_REGEX_TYPE) == "!":
                    text = match.group(TEXT_REGEX_TEXT)
                    self._texts.append((line_no, text.strip()))

        if symbol_attributes:
            self._symbols[symbol_attributes["InstName"]] = symbol_attributes

    def get_component_info(self, component) -> dict:
        """Returns the component information as a dictionary"""
        if component not in self._symbols:
            _logger.error(f"Component {component} not found in ASC file")
            raise ComponentNotFoundError(f"Component {component} not found in ASC file")
        return self._symbols[component]

    def _get_line_matching(self, command, search_expression: re.Pattern) -> Tuple[int, Union[re.Match, None]]:
        command_upped = command.upper()
        for line_no, line in self._texts:
            if line.upper().startswith(command_upped):
                match = search_expression.search(line)
                if match:
                    return line_no, match
        else:
            return -1, None

    def get_parameter(self, param: str) -> str:
        param_regex = re.compile(PARAM_REGEX % param, re.IGNORECASE)
        line_no, match = self._get_line_matching(".PARAM", param_regex)
        if match:
            return match.group('value')
        else:
            raise ParameterNotFoundError(f"Parameter {param} not found in ASC file")

    def set_parameter(self, param: str, value: Union[str, int, float]) -> None:
        param_regex = re.compile(PARAM_REGEX % param, re.IGNORECASE)
        line_no, match = self._get_line_matching(".PARAM", param_regex)
        if match:
            _logger.debug(f"Parameter {param} found in ASC file, updating it")
            if isinstance(value, (int, float)):
                value_str = format_eng(value)
            else:
                value_str = value
            line: str = self._asc_file_lines[line_no]
            match = param_regex.search(line)  # repeating the search, so we update the correct start/stop parameter
            start, stop = match.span(param_regex.groupindex['replace'])
            self._asc_file_lines[line_no] = line[:start] + "{}={}".format(param, value_str) + line[stop:]
            _logger.info(f"Parameter {param} updated to {value_str}")
            _logger.debug(f"Line:{line_no + 1} Updated to: {self._asc_file_lines[line_no]}")
        else:
            # Was not found so we need to add it,
            _logger.debug(f"Parameter {param} not found in ASC file, adding it")
            x, y = self._get_text_space()
            self._asc_file_lines.append("TEXT {} {} Left 2 !.param {}={}".format(x, y, param, value) + END_LINE_TERM)
            _logger.info(f"Parameter {param} added with value {value}")
            _logger.debug(f"Line:{len(self._asc_file_lines)} Added: {self._asc_file_lines[-1]}")
        self._parse_asc_file()

    def set_component_value(self, device: str, value: Union[str, int, float]) -> None:
        comp_info = self.get_component_info(device)
        start = comp_info['line']
        for offset, line in enumerate(self._asc_file_lines[start:]):
            if line.startswith("SYMATTR Value"):
                if isinstance(value, str):
                    value_str = value
                else:
                    value_str = format_eng(value)
                self._asc_file_lines[start + offset] = "SYMATTR Value {}".format(value_str) + END_LINE_TERM
                _logger.info(f"Component {device} updated to {value_str}")
                _logger.debug(f"Line:{start + offset + 1} Updated to: {self._asc_file_lines[start + offset]}")
                self._parse_asc_file()
                break
        else:
            _logger.error(f"Component {device} does not have a Value attribute")
            raise ComponentNotFoundError(f"Component {device} does not have a Value attribute")

    def set_element_model(self, element: str, model: str) -> None:
        comp_info = self.get_component_info(element)
        line_no = comp_info['line']
        tokens = self._asc_file_lines[line_no].split(' ')
        tokens[1] = model
        self._asc_file_lines[line_no] = ' '.join(tokens)
        _logger.info(f"Component {element} updated to {model}")

    def get_component_value(self, element: str) -> str:
        comp_info = self.get_component_info(element)
        if "Value" not in comp_info:
            _logger.error(f"Component {element} does not have a Value attribute")
            raise ComponentNotFoundError(f"Component {element} does not have a Value attribute")
        return comp_info["Value"]

    def get_components(self, prefixes='*') -> list:
        if prefixes == '*':
            return list(self._symbols.keys())
        return [k for k in self._symbols.keys() if k[0] in prefixes]

    def remove_component(self, designator: str):
        comp_info = self.get_component_info(designator)
        line_end = line_start = comp_info['line']
        for offset, line in enumerate(self._asc_file_lines[line_start:]):
            if line.startswith("SYMBOL") or line.startswith("WINDOW") or line.startswith("SYMATTR"):
                continue
            else:
                line_end = line_start + offset
                break  # If another line is found, then it is the start of another component
        del self._asc_file_lines[line_start:line_end]
        self._parse_asc_file()

    def _get_text_space(self):
        """
        Returns the coordinate on the Schematic File canvas where a text can be appended.
        """
        min_x = 100000  # High enough to be sure it will be replaced
        max_x = 0
        min_y = 100000  # High enough to be sure it will be replaced
        max_y = 0
        for line in self._asc_file_lines:
            if line.startswith("SHEET"):
                x, y = line.split()[2:4]
                min_x = int(x)
                min_y = int(y)
            elif line.startswith("WIRE"):
                x1, y1, x2, y2 = [int(x) for x in line.split()[1:5]]
                min_x = min(min_x, x1, x2)
                max_x = max(max_x, x1, x2)
                min_y = min(min_y, y1, y2)
                max_y = max(max_y, y1, y2)
            elif line.startswith("FLAG") or line.startswith("TEXT"):
                x1, y1 = [int(x) for x in line.split()[1:3]]
                min_x = min(min_x, x1)
                max_x = max(max_x, x1)
                min_y = min(min_y, y1)
                max_y = max(max_y, y1)
            elif line.startswith("SYMBOL"):
                x1, y1 = [int(x) for x in line.split()[2:4]]
                min_x = min(min_x, x1)
                max_x = max(max_x, x1)
                min_y = min(min_y, y1)
                max_y = max(max_y, y1)

        return min_x, max_y + 24  # Setting the text in the bottom left corner of the canvas

    def add_instruction(self, instruction: str) -> None:
        instruction = instruction.strip()  # Clean any end of line terminators
        command = instruction.split()[0].upper()

        if command in UNIQUE_SIMULATION_DOT_INSTRUCTIONS:
            # Before adding new instruction, if it is a unique instruction, we just replace it
            i = 0
            while i < len(self._texts):
                line_no, line = self._texts[i]
                command = line.split()[0].upper()
                if command in UNIQUE_SIMULATION_DOT_INSTRUCTIONS:
                    line = self._asc_file_lines[line_no]
                    self._asc_file_lines[line_no] = line[:line.find("!")] + instruction + END_LINE_TERM
                    self._parse_asc_file()
                    return  # Job done, can exit this method
                i += 1
        elif command.startswith('.PARAM'):
            raise RuntimeError('The .PARAM instruction should be added using the "set_parameter" method')
        # If we get here, then the instruction was not found, so we need to add it
        x, y = self._get_text_space()
        self._asc_file_lines.append("TEXT {} {} Left 2 !{}".format(x, y, instruction) + END_LINE_TERM)
        self._parse_asc_file()

    def remove_instruction(self, instruction: str) -> None:
        i = 0
        while i < len(self._texts):
            line_no, line = self._texts[i]
            if instruction in line:
                del self._asc_file_lines[line_no]
                _logger.info(f"Instruction {line} removed")
                self._parse_asc_file()
                return  # Job done, can exit this method
            i += 1

        msg = f'Instruction "{instruction}" not found'
        _logger.error(msg)
        raise RuntimeError(msg)

    def remove_Xinstruction(self, search_pattern: str) -> None:
        regex = re.compile(search_pattern, re.IGNORECASE)
        instr_removed = False
        i = 0
        while i < len(self._texts):
            line_no, line = self._texts[i]
            if regex.match(line):
                instr_removed = True
                del self._asc_file_lines[line_no]
                _logger.info(f"Instruction {line} removed")
                self._parse_asc_file()  # This is needed to update the self._texts list
            else:
                i += 1
        if not instr_removed:
            msg = f'Instructions matching "{search_pattern}" not found'
            _logger.error(msg)


#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        LTSteps.py
# Purpose:     Process LTSpice log files and align data for usage in a spreadsheet tool such as Excel, or Calc.
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

"""
This module allows to process data generated by LTSpice during simulation. There are three types of files that are
handled by this module.

    + log files - Files with the extension '.log' that are automatically generated during simulation, and that are
      normally accessible with the shortcut Ctrl+L after a simulation is ran.Log files are interesting for two reasons.

            1. If .STEP primitives are used, the log file contain the correspondence between the step run and the step
            value configuration.

            2. If .MEAS primitives are used in the schematic, the log file contains the measurements made on the output
            data.

      LTSteps.py can be used to retrieve both step and measurement information from log files.

    + txt files - Files exported from the Plot File -> Export data as text menu. This file is an text file where data is
      saved in the text format. The reason to use spicelib instead of another popular lib as pandas, is because the data
      format when .STEPS are used in the simulation is not not very practical. The spicelib LTSteps.py can be used to
      reformat the text, so that the run parameter is added to the data as an additional column instead of a table
      divider. Please Check LTSpiceExport class for more information.

    + mout files - Files generated by the Plot File -> Execute .MEAS Script menu. This command allows the user to run
      predefined .MEAS commands which create a .mout file. A .mout file has the measurement information stored in the
      following format:

      .. code-block:: text

            Measurement: Vout_rms
            step	RMS(V(OUT))	FROM	TO
             1	1.41109	0	0.001
             2	1.40729	0	0.001

            Measurement: Vin_rms
              step	RMS(V(IN))	FROM	TO
                 1	0.706221	0	0.001
                 2	0.704738	0	0.001

            Measurement: gain
              step	Vout_rms/Vin_rms
                 1	1.99809
                 2	1.99689


The LTSteps.py can be used directly from a command line by invoking python with the -m option as exemplified below.

.. code-block:: text

    $ python -m spicelib.LTSteps <path_to_filename>

If `<path_to_filename>` is a log file, it will create a file with the same name, but with extension .tout that is a
tab separated value (tsv) file, which contains the .STEP and .MEAS information collected.

If `<path_to_filename>` is a txt exported file, it will create a file with the same name, but with extension .tsv a
tab separated value (tsv) file, which contains data reformatted with the step number as one of the columns. Please
consult the reformat_LTSpice_export() function for more information.

If `<path_to_filename>` is a mout file, it will create a file with the same name, but with extension .tmout that is a
tab separated value (tsv) file, which contains the .MEAS information collected, but adding the STEP run information
as one of the columns.

If `<path_to_filename>` argument is ommited, the script will automatically search for the newest .log/.txt/.mout file
and use it.

"""
__author__ = "Nuno Canto Brum <me@nunobrum.com>"
__copyright__ = "Copyright 2023, Fribourg Switzerland"
import os
import sys

from .log.ltsteps import *
import logging
_logger = logging.getLogger("spicelib.LTSteps")


def main():
    """
    Main function for the LTSteps.py script
    """
    def valid_extension(filename):
        """A simple function to check if the filename has a valid extension"""
        return filename.endswith('.txt') or filename.endswith('.log') or filename.endswith('.mout')

    if len(sys.argv) > 1:
        filename = sys.argv[1]
        print("Using filename:", filename)
    else:
        # TODO: This should be changed to use the argparse module. The --last option should be added to indicate
        # that the last file should be used.
        filename = None
        newer_date = 0
        for f in os.listdir():
            date = os.path.getmtime(f)
            if date > newer_date and valid_extension(f):
                newer_date = date
                filename = f

    if filename is None:
        print("File not found")
        print("This tool only supports the following extensions :'.txt','.log','.mout'")
        exit(-1)
        
    if not valid_extension(filename):
        print("Invalid extension in filename '%s'" % filename)
        print("This tool only supports the following extensions :'.txt','.log','.mout'")
        exit(-1)

    fname_out = None
    if filename.endswith('.txt'):
        fname_out = filename[:-len('txt')] + 'tsv'
    elif filename.endswith('.log'):
        fname_out = filename[:-len('log')] + 'tlog'
    elif filename.endswith('.mout'):
        fname_out = filename[:-len('mout')] + 'tmout'
    else:
        print("Error in file type")
        print("This tool only supports the following extensions :'.txt','.log','.mout'")
        exit(-1)

    if fname_out is not None:
        print("Creating File %s" % fname_out)
        if filename.endswith('txt'):
            _logger.debug("Processing Data File")
            reformat_LTSpice_export(filename, fname_out)
        elif filename.endswith("log"):
            data = LTSpiceLogReader(filename)
            data.split_complex_values_on_datasets()
            data.export_data(fname_out)
        elif filename.endswith(".mout"):
            log_file = filename[:len('mout')] + 'log'
            if os.path.exists(log_file):
                steps = LTSpiceLogReader(log_file, read_measures=False)
                data = LTSpiceLogReader(filename, step_set=steps.stepset)
                data.stepset = steps.stepset
            else:
                # just reformats
                data = LTSpiceLogReader(filename)
            data.split_complex_values_on_datasets()
            data.export_data(fname_out)
            data.export_data(fname_out)


if __name__ == "__main__":
    _logger.setLevel(logging.INFO)
    main()



import os

from spicelib import SpiceEditor, SimRunner
from spicelib.simulators.ltspice_simulator import LTspice
from spicelib.sim.sim_stepping import SimStepper


def processing_data(raw_file, log_file):
    print("Handling the simulation data of %s" % log_file)

runner = SimRunner(parallel_sims=4, output_folder='./temp2', simulator=LTspice)

# select spice model
Stepper = SimStepper(SpiceEditor("./testfiles/Batch_Test.net"), runner)
# set default arguments

Stepper.set_parameters(res=0, cap=100e-6)
Stepper.set_component_value('R2', '2k')
Stepper.set_component_value('R1', '4k')
Stepper.set_element_model('V3', "SINE(0 1 3k 0 0 0)")
# define simulation
Stepper.add_instructions(
    "; Simulation settings",
    ";.param run = 0"
)
Stepper.set_parameter('run', 0)
Stepper.set_parameter('test_param2', 20)
Stepper.add_model_sweep('XU1', ('AD712', 'AD820'))
Stepper.add_value_sweep('V1', (5, 10, 15))
# Stepper.add_value_sweep('V1', (-5, -10, -15))

# run_netlist_file = "{}_{}_{}.net".format(Stepper.circuit_radic, opamp, supply_voltage)
Stepper.run_all(callback=processing_data)

# Sim Statistics
print('Successful/Total Simulations: ' + str(Stepper.okSim) + '/' + str(Stepper.runno))
runner.file_cleanup()



from spicelib.simulators.qspice_simulator import Qspice
from spicelib.sim.simulator import run_function

qpost = [Qspice.spice_exe[0].replace("QSPICE64.exe", "QPOST.exe")]

run_function(qpost, "C:\\Users\\nuno\\Documents\\QSpice\\examples\\test.qraw", "- o C:\\Users\\nuno\\Documents\\QSpice\\examples\\test.meas")



import numpy as np
from spicelib import RawRead, Trace, RawWrite


def test_readme_snippet():
    LW = RawWrite(fastacces=False)
    tx = Trace('time', np.arange(0.0, 3e-3, 997E-11))
    vy = Trace('N001', np.sin(2 * np.pi * tx.data * 10000))
    vz = Trace('N002', np.cos(2 * np.pi * tx.data * 9970))
    LW.add_trace(tx)
    LW.add_trace(vy)
    LW.add_trace(vz)
    LW.save("./testfiles/teste_snippet1.raw")


def test_trc2raw():  # Convert Teledyne-Lecroy trace files to raw files
    f = open(r"./testfiles/Current_Lock_Front_Right_8V.trc")
    raw_type = ''  # Initialization of parameters that need to be overridden by the file header
    wave_size = 0
    for line in f:
        tokens = line.rstrip('\r\n').split(',')
        if len(tokens) == 4:
            if tokens[0] == 'Segments' and tokens[2] == 'SegmentSize':
                wave_size = int(tokens[1]) * int(tokens[3])
        if len(tokens) == 2:
            if tokens[0] == 'Time' and tokens[1] == 'Ampl':
                raw_type = 'transient'
                break
    if raw_type == 'transient' and wave_size > 0:
        data = np.genfromtxt(f, dtype='float,float', delimiter=',', max_rows=wave_size)
        LW = RawWrite()
        LW.add_trace(Trace('time', [x[0] for x in data]))
        LW.add_trace(Trace('Ampl', [x[1] for x in data]))
        LW.save("teste_trc.raw")
    f.close()


def test_axis_sync():  # Test axis sync
    LW = RawWrite()
    tx = Trace('time', np.arange(0.0, 3e-3, 997E-11))
    vy = Trace('N001', np.sin(2 * np.pi * tx.data * 10000))
    vz = Trace('N002', np.cos(2 * np.pi * tx.data * 9970))
    LW.add_trace(tx)
    LW.add_trace(vy)
    LW.add_trace(vz)
    LW.save("./testfiles/teste_w.raw")
    LR = RawRead("./testfiles/testfile.raw")
    LW.add_traces_from_raw(LR, ('V(out)',), force_axis_alignment=True)
    LW.save("./testfiles/merge.raw")
    test = """
    equal = True
    for ii in range(len(tx)):
        if t[ii] != tx[ii]:
            print(t[ii], tx[ii])
            equal = False
    print(equal)

    v = LR.get_trace('N001')
    max_error = 1.5e-12
    for ii in range(len(vy)):
        err = abs(v[ii] - vy[ii])
        if err > max_error:
            max_error = err
            print(v[ii], vy[ii], v[ii] - vy[ii])
    print(max_error)
    """


def test_write_ac():
    LW = RawWrite()
    LR = RawRead("./testfiles/PI_Filter.raw")
    LR1 = RawRead("./testfiles/PI_Filter_resampled.raw")
    LW.add_traces_from_raw(LR, ('V(N002)',))
    LW.add_traces_from_raw(LR1, 'V(N002)', rename_format='N002_resampled', force_axis_alignment=True)
    LW.flag_fastaccess = False
    LW.save("./testfiles/PI_filter_rewritten.raw")
    LW.flag_fastaccess = True
    LW.save("./testfiles/PI_filter_rewritten_fast.raw")


def test_write_tran():
    LR = RawRead("./testfiles/TRAN - STEP.raw")
    LW = RawWrite()
    LW.add_traces_from_raw(LR, ('V(out)', 'I(C1)'))
    LW.flag_fastaccess = False
    LW.save("./testfiles/TRAN - STEP0_normal.raw")
    LW.flag_fastaccess = True
    LW.save("./testfiles/TRAN - STEP0_fast.raw")


def test_combine_tran():
    LW = RawWrite()
    for tag, raw in (
            ("AD820_15", "./testfiles/Batch_Test_AD820_15.raw"),
            # ("AD820_10", "./testfiles/Batch_Test_AD820_10.raw"),
            ("AD712_15", "./testfiles/Batch_Test_AD712_15.raw"),
            # ("AD712_10", "./testfiles/Batch_Test_AD712_10.raw"),
            # ("AD820_5", "./testfiles/Batch_Test_AD820_5.raw"),
            # ("AD712_5", "./testfiles/Batch_Test_AD712_5.raw"),
    ):
        LR = RawRead(raw)
        LW.add_traces_from_raw(LR, ("V(out)", "I(R1)"), rename_format="{}_{tag}", tag=tag, force_axis_alignment=True)
    LW.flag_fastaccess = False
    LW.save("./testfiles/Batch_Test_Combine.raw")


test_readme_snippet()
test_axis_sync()
test_write_ac()
test_write_tran()
test_combine_tran()



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        quick_sensitivity_analysis.py
# Purpose:     Classes to make a sensitivity analysis
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     16-10-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
from typing import Union, Optional

from .tolerance_deviations import ToleranceDeviations, DeviationType
from ..sim_runner import AnyRunner
from ...editor.base_editor import BaseEditor
from ...log.logfile_data import LogfileData


class QuickSensitivityAnalysis(ToleranceDeviations):
    """Class to automate Sensitivity simulations"""
    def __init__(self, circuit_file: Union[str, BaseEditor], runner: Optional[AnyRunner] = None):
        super().__init__(circuit_file, runner)
        self.components_analysed = []

    def prepare_testbench(self, **kwargs):
        """Prepares the simulation by setting the tolerances for each component"""
        no = 0
        self.components_analysed.clear()
        use_min = kwargs.get('use_min', False)
        for comp in self.get_components('*'):
            val, dev = self.get_component_value_deviation_type(comp)
            new_val = val
            if dev.typ == DeviationType.tolerance:
                new_val = "{satol(%s,%g,%d)}" % (val, dev.max_val, no)
            elif dev.typ == DeviationType.minmax:
                used_value = dev.min_val if use_min else dev.max_val
                new_val = "{sammx(%s,%g,%d)}" % (val, used_value, no)

            if new_val != val:
                self.set_component_value(comp, new_val)
                self.components_analysed.append(comp)
                no += 1

        self.num_runs = no - 1
        if use_min:
            self.editor.add_instruction(".func satol(nom,tol,idx) nom*if(run==idx,1-tol,1)")
        else:
            self.editor.add_instruction(".func satol(nom,tol,idx) nom*if(run==idx,1+tol,1)")
        self.editor.add_instruction(".func sammx(nom,val,idx) if(run==idx,val,nom)")
        self.editor.add_instruction(".step param run -1 %d 1" % self.num_runs)
        self.editor.set_parameter('run', -1)  # in case the step is commented.
        self.testbench_prepared = True

    def get_sensitivity_data(self, ref: str, measure: str) -> Union[float, dict]:
        """
        Returns the sensitivity data for a given component and measurement in terms of percentage of the total error.
        This quick approach is not very accurate, but it is fast. It assumes that the system is linear and that the
        maximum error is the sum of the absolute error of each component. This is a rough approximation, but it is
        good enough for a quick analysis. For more accurate results, use the Worst Case Analysis, which requires
        more simulation runs but gives a more accurate result.
        The best compromise, is to start with the quick analysis and then use the Worst Case Analysis to refine the
        results with only the components that have a significant contribution to the error.
        :param ref: The reference component, or '*' to return a dictionary with all the components
        :param measure: The measurement to be analysed
        :return: The sensitivity data in percentage of the total error for the reference component
        """
        if self.testbench_prepared and self.testbench_executed:
            log_data: LogfileData = self.read_logfiles()
            nominal_data = log_data.get_measure_value(measure, run=-1)
            error_data = []
            for idx, _ in enumerate(self.components_analysed):
                step_data = log_data.get_measure_value(measure, run=idx)
                error_data.append(abs(step_data - nominal_data))
            total_error = sum(error_data)
            if ref == '*':
                return {ref: error_data[idx] / total_error * 100 if total_error != 0 else 0
                        for idx, ref in enumerate(self.components_analysed)}
            else:
                idx = self.components_analysed.index(ref)
                return error_data[idx] / total_error * 100 if total_error != 0 else 0
        else:
            raise RuntimeError("Testbench not prepared or executed")

    def run_analysis(self):
        pass


#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        failure_modes.py
# Purpose:     Class to automate FMEA
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     10-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

from collections import OrderedDict
from typing import Union, Optional, Iterable, Type

from ..simulator import Simulator
from ...editor.base_editor import BaseEditor, ComponentNotFoundError
from .sim_analysis import SimAnalysis, AnyRunner


class FailureMode(SimAnalysis):
    """
    This Class will replace each component on the circuit for their failure modes and launch a simulation.

    The following failure modes are built-in:

        * Resistors, Capacitors, Inductors and Diodes
            # Open Circuit
            # Short Circuit

        * Transistors
            # Open Circuit (All pins)
            # Short Circuit (All pins)
            # Short Circuit Base-Emitter (Bipolar) / Gate-Source (MOS)
            # Short Circuit Collector-Emitter (Bipolar) / Drain-Source (MOS)

        * Integrated Circuits
            # The failure modes are defined by the user by using the add_failure_mode() method
    """
    def __init__(self, circuit_file: Union[str, BaseEditor], simulator: Optional[Type[Simulator]] = None,
                 runner: Optional[AnyRunner] = None):
        SimAnalysis.__init__(self, circuit_file, simulator, runner)
        self.resistors = self.editor.get_components('R')
        self.capacitors = self.editor.get_components('C')
        self.inductors = self.editor.get_components('L')
        self.diodes = self.editor.get_components('D')
        self.bipolars = self.editor.get_components('Q')
        self.mosfets = self.editor.get_components('M')
        self.subcircuits = self.editor.get_components('X')
        self.user_failure_modes = OrderedDict()

    def add_failure_circuit(self, component, sub_circuit):
        if not component.startswith('X'):
            raise RuntimeError("The failure modes addition only works with sub circuits")
        if component not in self.subcircuits:
            raise ComponentNotFoundError()
        raise NotImplementedError("TODO")  # TODO: Implement this

    def add_failure_mode(self, component, short_pins: Iterable, open_pins: Iterable):
        if not component.startswith('X'):
            raise RuntimeError("The failure modes addition only works with subcircuits")
        if component not in self.subcircuits:
            raise ComponentNotFoundError()
        raise NotImplementedError("TODO")  # TODO: Implement this

    def run_all(self):
        for resistor in self.resistors:
            # Short Circuit
            self.editor.set_component_value(resistor, '1f')  # replaces the resistor with a one femto-Ohm
            self.simulations[f"{resistor}_S"] = self.runner.run()
            # Open Circuit
            self.editor.remove_component(resistor)
            self.simulations[f"{resistor}_O"] = self.simulator.run()
            self.editor.reset_netlist()

        for two_pin_comps in (self.capacitors, self.inductors, self.diodes):
            for two_pin_component in two_pin_comps:
                # Open Circuit
                cinfo = self.editor.get_component_info(two_pin_component)
                self.editor.remove_component(two_pin_component)
                self.simulations[f"{two_pin_component}_O"] = self.simulator.run()
                # Short Circuit
                self.editor.netlist[cinfo['line']] = f"Rfmea_short_{two_pin_component}{cinfo['nodes']} 1f"
                self.simulations[f"{two_pin_component}_S"] = self.simulator.run()
                self.editor.reset_netlist()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        xyce_simulator.py
# Purpose:     Tool used to launch NGspice simulations in batch mode.
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     14-03-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

import sys
import os

from pathlib import Path
from typing import Union
import logging
_logger = logging.getLogger("spicelib.XYCESimulator")
from ..sim.simulator import Simulator, run_function


class XyceSimulator(Simulator):
    """Stores the simulator location and command line options and runs simulations."""
    spice_exe = ["C:/Program Files/Xyce 7.6 NORAD/bin/xyce.exe"]
    process_name = "XVIIx64.exe"
    xyce_args = {
        '-b'                : ['-b'],  # batch mode flag for spice compatibility (ignored)
        '-h'                : ['-h'],  # print usage and exit
        '-v'                : ['-v'],  # print version info and exit
        '-capabilities'     : ['-capabilities'],  # print compiled-in options and exit
        '-license'          : ['-license'],  # print license and exit
        '-param'            : ['-param', '<param_options>'],
        # [device [level [<inst|mod>]]] print a terse summary of model and/or device parameters
        '-doc'              : ['-doc', '<param_options>'],
        # [device [level [<inst|mod>]]] output latex tables of model and device parameters to files
        '-doc_cat'          : ['-doc_cat', '<param_options>'],
        # [device [level [<inst|mod>]]] output latex tables of model and device parameters to files
        '-count'            : ['-count'],  # device count without netlist syntax or topology check
        '-syntax'           : ['-syntax'],  # check netlist syntax and exit
        '-norun'            : ['-norun'],  # netlist syntax and topology and exit
        '-namesfile'        : ['-namesfile', '<path>'],  # output internal names file to <path> and exit
        '-noise_names_file' : ['-noise_names_file', '<path>'],  # output noise source names file to <path> and exit
        '-quiet'            : ['-quiet'],  # suppress some of the simulation-progress messages sent to stdout
        '-jacobian_test'    : ['-jacobian_test'],  # jacobian matrix diagnostic
        '-hspice-ext'       : ['-hspice-ext', '<hsext_options>'],
        # apply hspice compatibility features during parsing.  option=all applies them all
        '-redefined_params' : ['-redefined_params', '<redef_param_option>'],
        # set option for redefined .params as ignore (use last), usefirst, warn or error
        '-subckt_multiplier': ['-subckt_multiplier', '<truefalse_option>'],
        # set option to true(default) or false to apply implicit subcircuit multipliers
        '-delim'            : ['-delim', '<delim_option>'],  # <TAB|COMMA|string>   set the output file field delimiter
        '-o'                : ['-o', '<basename>'],  # <basename> for the output file(s)
        '-l'                : ['-l', '<path>'],  # place the log output into <path>, "cout" to log to stdout
        '-per-processor'    : ['-per-processor'],  # create log file for each processor, add .<n>.<r> to log path
        '-remeasure'        : ['-remeasure', '<path>'],
        # [existing Xyce output file] recompute .measure() results with existing data
        '-nox'              : ['-nox', 'onoff_option'],  # <on|off>               NOX nonlinear solver usage
        '-linsolv'          : ['-linsolv', '<solver>'],  # <solver>           force usage of specific linear solver
        '-maxord'           : ['-maxord', '<int_option>'],  # <1..5>              maximum time integration order
        '-max'              : ['-max', '<int_option>'],  # -warnings <#>           maximum number of warning messages
        '-prf'              : ['-prf', '<path>'],  # <param file name>      specify a file with simulation parameters
        '-rsf'              : ['-rsf', '<path>'],  # specify a file to save simulation responses functions.
        '-r'                : ['-r', '<path>'],  # <file>   generate a rawfile named <file> in binary format
        '-a'                : ['-a', '<path>'],  # use with -r <file> to output in ascii format
        '-randseed'         : ['-randseed', '<int_option>'],
        # <number>          seed random number generator used by expressions and sampling methods
    }

    @classmethod
    def valid_switch(cls, switch, parameter='') -> list:
        """
        Validates a command line switch. The following options are available for Xyce:

        :param switch: switch to be added. If the switch is not on the list above, it should be correctly formatted with
            the preceding '-' switch
        :type switch: str
        :param parameter: parameter for the switch
        :type parameter: str, optional
        :return: the correct formatting for the switch
        :rtype: list
        """
        ret = []  # This is an empty switch
        if switch in cls.xyce_args:
            switch_list = cls.xyce_args[switch]
            if len(switch_list) == 2:
                param_token = switch_list[1]
                if param_token == '<path>':
                    ret = [switch_list[0], parameter]
                elif param_token == '<param_options>':
                    # Check for [device [level [<inst|mod>]]] syntax ??
                    ret = [switch_list[0], parameter]
                elif param_token == '<hsext_options>':
                    ret = [switch_list[0], parameter]
                elif param_token == '<redef_param_option>':
                    if parameter in ('ignore', 'uselast', 'usefirst', 'warn', 'error'):
                        ret = [switch_list[0], parameter]
                elif param_token == '<truefalse_option>':
                    if parameter.lower() in ('true', 'false'):
                        ret = [switch_list[0], parameter]
                elif param_token == '<delim_option>':
                    ret = [switch_list[0], parameter]
                elif param_token == '<onoff_option>':
                    if parameter.lower() in ('on', 'off'):
                        ret = [switch_list[0], parameter]
                elif param_token == '<int_option>':
                    try:
                        int(parameter)
                    except ValueError:
                        pass
                    else:
                        ret = [switch_list[0], parameter]
                else:
                    _logger.warning(f"Invalid parameter {parameter} for switch '{switch}'")
            else:
                ret = switch_list
        else:
            _logger.error(f"Invalid Switch {switch}")
        return ret

    @classmethod
    def run(cls, netlist_file, cmd_line_switches, timeout):
        cmd_run = cls.spice_exe + cmd_line_switches + [netlist_file]
        # start execution
        return run_function(cmd_run, timeout=timeout)



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        montecarlo.py
# Purpose:     Classes to automate Monte-Carlo simulations
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     10-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

import random
import logging
from typing import Union, Callable, Type

_logger = logging.getLogger("spicelib.SimAnalysis")

from .tolerance_deviations import ToleranceDeviations, DeviationType, ComponentDeviation
from ..process_callback import ProcessCallback
from ...log.logfile_data import LogfileData



class Montecarlo(ToleranceDeviations):
    """Class to automate Monte-Carlo simulations"""

    def prepare_testbench(self, **kwargs):
        """
        Prepares the simulation by setting the tolerances for the components
        :keyword num_runs: Number of runs to be performed. Default is 1000.
        :return: None
        """
        min_max_uni_func = False
        min_max_norm_func = False
        tol_uni_func = False
        tol_norm_func = False
        for ref in self.get_components('*'):
            val, dev = self.get_component_value_deviation_type(ref)  # get there present value
            new_val = val
            if dev.typ == DeviationType.tolerance:
                tolstr = ('%g' % dev.max_val).rstrip('0').rstrip('.')
                if dev.distribution == 'uniform':
                    new_val = "{utol(%s,%s)}" % (val, tolstr)  # calculate expression for new value
                    tol_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{ntol(%s,%s)}" % (val, tolstr)
                    tol_norm_func = True
            elif dev.typ == DeviationType.minmax:
                if dev.distribution == 'uniform':
                    new_val = "{urng(%s, %s,%s)}" % (val, dev.min_val, dev.max_val)  # calculate expression for new value
                    min_max_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{nrng(%s,%s,%s)}" % (val, dev.min_val, dev.max_val)
                    min_max_norm_func = True

            if new_val != val:  # Only update the value if it has changed
                self.set_component_value(ref, new_val)  # update the value

        for param in self.parameter_deviations:
            val, dev = self.get_parameter_value_deviation_type(param)
            new_val = val
            if dev.typ == DeviationType.tolerance:
                if dev.distribution == 'uniform':
                    new_val = "{utol(%s,%g)}" % (val, dev.max_val)
                    tol_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{ntol(%g,%g)}" % (val, dev.max_val)
                    tol_norm_func = True
            elif dev.typ == DeviationType.minmax:
                if dev.distribution == 'uniform':
                    new_val = "{urng(%s,%g,%g)}" % (val, (dev.max_val+dev.min_val)/2, (dev.max_val-dev.min_val)/2)
                    min_max_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{nrng(%s,%g,%g)}" % (val, (dev.max_val+dev.min_val)/2, (dev.max_val-dev.min_val)/6)
                    min_max_norm_func = True
            else:
                continue
            self.editor.set_parameter(param, new_val)

        if tol_uni_func:
            self.editor.add_instruction(".func utol(nom,tol) if(run<0, nom, mc(nom,tol))")

        if tol_norm_func:
            self.editor.add_instruction(".func ntol(nom,tol) if(run<0, nom, nom*(1+gauss(tol/3)))")

        if min_max_uni_func:
            self.editor.add_instruction(".func urng(nom,mean,df2) if(run<0, nom, mean*flat(df2))")

        if min_max_norm_func:
            self.editor.add_instruction(".func nrng(nom,mean,df6) if(run<0, nom, mean*(1+gauss(df6)))")

        self.num_runs = kwargs.get('num_runs', self.num_runs if self.num_runs != 0 else 1000)
        self.editor.add_instruction(".step param run -1 %d 1" % self.num_runs)
        self.editor.set_parameter('run', -1)
        self.testbench_prepared = True

    def _get_sim_value(self, value: float, dev: ComponentDeviation) -> float:
        """Returns a new value for the simulation"""
        new_val = value
        if dev.typ == DeviationType.tolerance:
            if dev.distribution == 'uniform':
                new_val = f"{random.Random().uniform(value * (1- dev.max_val), value * (1 + dev.max_val)):g}"
            elif dev.distribution == 'normal':
                new_val = f"{random.Random().gauss(value, dev.max_val / 3):g}"
        elif dev.typ == DeviationType.minmax:
            if dev.distribution == 'uniform':
                new_val = f"{random.Random().uniform(dev.min_val, dev.max_val):g}"
            elif dev.distribution == 'normal':
                new_val = f"{random.Random().gauss((dev.max_val + dev.min_val) / 2, (dev.max_val - dev.min_val) / 6):g}"
        else:
            _logger.warning("Unknown deviation type")
        return new_val

    def run_analysis(self, num_runs: int = 1000,
                     callback: Union[Type[ProcessCallback], Callable] = None,
                     callback_args: Union[tuple, dict] = None,
                     switches=None,
                     timeout: float = None,
                     ):
        """This method runs the analysis without updating the netlist.
        It will update component values and parameters according to their deviation type and call the simulation.
        The advantage of this method is that it doesn't require adding random functions to the netlist.
        The number of times the simulation is done is specified on the argument num_runs."""
        self.clear_simulation_data()
        for run in range(num_runs):
            self._reset_netlist() # reset the netlist
            self.play_instructions() # play the instructions
            # Preparing the variation on components
            for ref in self.get_components('*'):
                val, dev = self.get_component_value_deviation_type(ref)  # get there present value
                new_val = self._get_sim_value(val, dev)
                if new_val != val:  # Only update the value if it has changed
                    self.editor.set_component_value(ref, new_val)  # update the value
            # Preparing the variation on parameters
            for param in self.parameter_deviations:
                val, dev = self.get_parameter_value_deviation_type(param)
                new_val = self._get_sim_value(val, dev)
                if new_val != val:  # Only update the value if it has changed
                    self.editor.set_parameter(param, new_val)
            # Run the simulation
            rt = self.run(self.editor, wait_resource=True,
                          callback=callback, callback_args=callback_args,
                          switches=switches, timeout=timeout)

        self.runner.wait_completion()
        if callback is not None:
            callback_rets = []
            for rt in self.run_tasks:
                callback_rets.append(rt.get_results())
            self.simulation_results['callback_returns'] = callback_rets

    def analyse_measurement(self, meas_name: str):
        if 'log_data' not in self.simulation_results:
            log_data = self.read_logfiles()
            self.simulation_results['log_data'] = log_data

        log_data: LogfileData = self.simulation_results['log_data']
        meas_data = log_data[meas_name]
        if meas_data is None:
            _logger.warning("Measurement %s not found in log files", meas_name)
            return None
        else:
            return meas_data



from spicelib import SpiceEditor, SimRunner, RawRead, LTSteps
from spicelib.simulators.ltspice_simulator import LTspice
from pathlib import Path

# You should replace it by your own executor path.

output_folder = "./py_sim_out"

runner = SimRunner(output_folder=output_folder, parallel_sims=16, simulator=LTspice)

for time in range(1, 200, 1):
    netlist = SpiceEditor('stuck_template.net')
    runner.run(netlist, run_filename=f'test{time}.net')

runner.wait_completion()


from spicelib import RawRead
from spicelib.sim.sim_runner import SimRunner
from spicelib.editor.spice_editor import SpiceEditor
from spicelib.simulators.qspice_simulator import Qspice
from spicelib.utils.sweep_iterators import sweep_log


def processing_data(raw_file, log_file):
    print("Handling the simulation data of %s, log file %s" % (raw_file, log_file))
    raw_data = RawRead(raw_file)
    vout = raw_data.get_wave('V(out)')
    return raw_file, vout.max()


# select spice model
sim = SimRunner(output_folder='./temp', simulator=Qspice.create_from('C:/Program Files/QSPICE/QSPICE64.exe'))
netlist = SpiceEditor('./testfiles/testfile.net')
# set default arguments
netlist.set_component_value('R1', '4k')
netlist.set_element_model('V1', "SINE(0 1 3k 0 0 0)")  # Modifying the
netlist.add_instruction(".tran 1n 3m")
netlist.add_instruction(".plot V(out)")
netlist.add_instruction(".save all")

sim_no = 1
# .step dec param cap 1p 10u 1
for cap in sweep_log(1e-12, 10e-6, 10):
    netlist.set_component_value('C1', cap)
    sim.run(netlist, callback=processing_data, run_filename=f'testfile_qspice_{sim_no}.net')
    sim_no += 1

# Reading the data
results = {}
for raw_file, vout_max in sim:  # Iterate over the results of the callback function
    results[raw_file.name] = vout_max
# The block above can be replaced by the following line
# results = {raw_file.name: vout_max for raw_file, vout_max in sim}

print(results)

# Sim Statistics
print('Successful/Total Simulations: ' + str(sim.okSim) + '/' + str(sim.runno))

sim.file_cleanup()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sim_server.py
# Purpose:     A simulation server that can execute simulations by request of a client located in a different machine.
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-02-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
from typing import Tuple
from xmlrpc.client import Binary
from xmlrpc.server import SimpleXMLRPCServer
import logging
_logger = logging.getLogger("spicelib.SimServer")

import threading
from pathlib import Path
import zipfile
import io
from spicelib.client_server.srv_sim_runner import ServerSimRunner
import uuid


class SimServer():

    def __init__(self, simulator, parallel_sims=4, output_folder='./temp', port=9000):
        self.output_folder = output_folder
        self.simulation_manager = ServerSimRunner(parallel_sims=parallel_sims, timeout=5 * 60, verbose=False,
                                                  output_folder=output_folder, simulator=simulator)
        self.server = SimpleXMLRPCServer(
                ('localhost', port),
                # requestHandler=RequestHandler
        )
        self.server.register_introspection_functions()
        self.server.register_instance(self)
        self.sessions = {}  # this will contain the session_id ids hashing their respective list of sim_tasks
        self.simulation_manager.start()
        self.server_thread = threading.Thread(target=self.server.serve_forever, name="ServerThread")
        self.server_thread.start()

    def run(self, session_id, circuit_name, zip_data):
        _logger.info(f"Run {session_id} : {circuit_name}")
        if session_id not in self.sessions:
            return -1  # This indicates that no job is started
        # Create a buffer from the zip data
        zip_buffer = io.BytesIO(zip_data.data)
        _logger.debug("Created the buffer")
        # Extract the contents of the zip file
        with zipfile.ZipFile(zip_buffer, 'r') as zip_file:
            _logger.debug(zip_file.namelist())
            zip_file.extract(circuit_name, self.output_folder)

        circuit_name = Path(self.output_folder) / circuit_name
        _logger.info(f"Running simulation of {circuit_name}")
        runno = self.simulation_manager.add_simulation(circuit_name)
        if runno != -1:
            self.sessions[session_id].append(runno)
        return runno

    def start_session(self):
        """Returns an unique key that represents the session. It will be later used to sort the sim_tasks belonging
        to the session."""
        session_id = str(uuid.uuid4())  # Needs to be a string, otherwise the rpc client can't handle it
        _logger.info(f"Starting session {session_id}")
        self.sessions[session_id] = []
        return session_id

    def status(self, session_id):
        """
        Returns a dictionary with task information. The key for the dictionary is the simulation identifier returned
        by the simulation start command. The value associated with each simulation identifier is another dictionary
        containing the following keys:

            * 'completed' - whether the simulation is already finished

            * 'start' - time when the simulation was started

            * 'stop' - server time
        """
        _logger.debug(f"collecting status for {session_id}")
        ret = []
        for task_info in self.simulation_manager.completed_tasks:
            _logger.debug(task_info)
            runno = task_info['runno']
            if runno in self.sessions[session_id]:
                ret.append(runno)  # transfers the dictionary from the simulation_manager completed task
                # to the return dictionary 
        _logger.debug(f"returning status {ret}")
        return ret

    def get_files(self, session_id, runno) -> Tuple[str, Binary]:
        if runno in self.sessions[session_id]:

            for task_info in self.simulation_manager.completed_tasks:
                if runno == task_info['runno']:
                    # Create a buffer to store the zip file in memory
                    zip_file = task_info['zipfile']
                    zip = zip_file.open('rb')
                    # Read the zip file from the buffer and send it to the server
                    zip_data = zip.read()
                    zip.close()
                    self.simulation_manager.erase_files_of_runno(runno)
                    return zip_file.name, Binary(zip_data)

        return "", Binary(b'')  # Returns and empty data

    def close_session(self, session_id):
        """Cleans all the pending sim_tasks with """
        _logger.info(f"Closing session {session_id}")
        for runno in self.sessions[session_id]:
            self.simulation_manager.erase_files_of_runno(runno)
        return True  # Needs to return always something. None is not supported

    def stop_server(self):
        _logger.debug("stopping...ServerInterface")
        self.simulation_manager.stop()
        self.server.shutdown()
        _logger.info("stopped...ServerInterface")
        return True  # Needs to return always something. None is not supported

    def running(self):
        return self.simulation_manager.running()




# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sweep_iterators.py
# Purpose:     Iterators to use for sweeping values
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     24-07-2020
# Licence:     refer to the LICENSE file
#
# -------------------------------------------------------------------------------

import math
from typing import Union, Optional, Iterable

__all__ = ['sweep', 'sweep_n', 'sweep_log', 'sweep_log_n']

class BaseIterator(object):
    """Common implementation to all Iterator classes"""

    def __init__(self, start: Union[int, float], stop: Optional[Union[int, float]] = None, step: Union[int, float] = 1):
        if stop is None:
            self.stop = start
            self.start = 0
        else:
            self.start = start
            self.stop = stop
        self.step = step
        self.finished = False

    def __iter__(self):
        self.finished = False
        return self

    def __next__(self):
        raise NotImplemented("This function needs to be overriden")


class sweep(BaseIterator):
    """
    Generator function to be used in sweeps.
    Advantages towards the range python built-in functions
    - Supports floating point arguments
    - Supports both up and down sweeps
    Usage:
        >>> list(sweep(0.3, 1.1, 0.2))
        [0.3, 0.5, 0.7, 0.9000000000000001, 1.1]
        >>> list(sweep(15, -15, 2.5))
        [15, 12.5, 10.0, 7.5, 5.0, 2.5, 0.0, -2.5, -5.0, -7.5, -10.0, -12.5, -15.0]
    """
    def __init__(self, start: Union[int, float], stop: Optional[Union[int, float]] = None,
                 step: Union[int, float] = 1):
        super().__init__(start, stop, step)
        assert step != 0, "Step cannot be 0"
        if self.step < 0 and self.start < self.stop:
            # The sign of the step determines whether it counts up or down.
            self.start, self.stop = self.stop, self.start
        elif self.step > 0 and self.stop < self.start:
            self.step = - self.step  # In this case invert the sigh
        self.niter = 0

    def __iter__(self):
        super().__iter__()
        # Resets the interator
        self.niter = 0
        return self

    def __next__(self):
        val = self.start + self.niter * self.step
        self.niter += 1
        if (self.step > 0 and val <= self.stop) or (self.step < 0 and val >= self.stop):
            return val
        else:
            self.finished = True
            raise StopIteration

def sweep_n(start: Union[int, float], stop: Union[int, float], N: int) -> Iterable[float]:
    """Helper function.
    Generator function that generates a 'N' number of points between a start and a stop interval.
    Advantages towards the range python built-in functions
    - Supports floating point arguments
    - Supports both up and down sweeps-
    Usage:
        >>> list(sweep_n(0.3, 1.1, 5))
        [0.3, 0.5, 0.7, 0.9000000000000001, 1.1]
        >>> list(sweep_n(15, -15, 13))
        [15, 12.5, 10.0, 7.5, 5.0, 2.5, 0.0, -2.5, -5.0, -7.5, -10.0, -12.5, -15.0]
        """
    return sweep(start, stop, (stop-start)/(N-1))



class sweep_log(BaseIterator):
    """
    Generator function to be used in logarithmic sweeps.
    Advantages towards the range python built-in functions_
    - Supports floating point arguments
    - Supports both up and down sweeps.
    Usage:
        >>> list(sweep_log(0.1, 11e3, 10))
        [0.1, 1.0, 10.0, 100.0, 1000.0, 10000.0]
        >>> list(sweep_log(1000, 1, 2))
        [1000, 500.0, 250.0, 125.0, 62.5, 31.25, 15.625, 7.8125, 3.90625, 1.953125]
    """
    def __init__(self, start: Union[int, float], stop: Optional[Union[int, float]] = None,
                 step: Optional[Union[int, float]] = 10):
        if stop is None:
            stop = start
            start = 1
        super().__init__(start, stop, step)
        assert step != 1 and step > 0, "Step must be higher than 0 and not 1"
        if self.start < self.stop and self.step < 1:
            self.start, self.stop = self.stop, self.start
        elif self.stop < self.start and self.step > 1:
            self.step = 1/self.step
        self.val = self.start

    def __iter__(self):
        super().__iter__()
        self.val = self. start
        return self

    def __next__(self):
        val = self.val  # Store previous value
        self.val *= self.step  # Calculate the next item
        if (self.start < self.stop and val <= self.stop) or \
           (self.start > self.stop and val >= self.stop):
            return val
        else:
            self.finished = True
            raise StopIteration


class sweep_log_n(BaseIterator):
    """Helper function.
    Generator function that generates a 'N' number of points between a start and a stop interval.
    Advantages towards the range python built-in functions
    - Supports floating point arguments
    - Supports both up and down sweeps-
    Usage:
        >>> list(sweep_log_n(1, 10, 6))
        [1.0, 1.5848931924611136, 2.5118864315095806, 3.9810717055349736, 6.309573444801934, 10.000000000000004]
        >>> list(sweep_log_n(10, 1, 5))
        [1.0, 0.5623413251903491, 0.31622776601683794, 0.17782794100389226, 0.09999999999999999]
        """
    def __init__(self, start: Union[int, float], stop: Optional[Union[int, float]], number_of_elements: int):
        step = math.exp(math.log(stop / start) / (number_of_elements - 1))
        assert step != 0, "Step cannot be 0"
        super().__init__(start, number_of_elements, step)
        self.niter = 0

    def __iter__(self):
        super().__iter__()
        self.niter = 0
        return self

    def __next__(self):
        if self.niter < self.stop:
            val = self.start * (self.step ** self.niter)
            self.niter += 1
            return val
        else:
            self.finished = True
            raise StopIteration


if __name__ == "__main__":

    print("list(sweep(10))", list(sweep(10)))
    print("list(sweep(1, 8))", list(sweep(1, 8)))
    print("list(sweep(2, 8, 2))", list(sweep(2, 8, 2)))
    print("list(sweep(2, 8, -2))", list(sweep(2, 8, -2)))
    print("list(sweep(8, 2, 2))", list(sweep(8, 2, 2)))
    print("list(sweep(0.3, 1.1, 0.2))", list(sweep(0.3, 1.1, 0.2)))
    print("list(sweep(15, -15, 2.5))", list(sweep(15, -15, 2.5)))
    print("list(sweep(-2, 2, 2))", list(sweep(-2, 2, 2)))
    print("list(sweep(-2, 2, -2))", list(sweep(-2, 2, -2)))
    print("list(sweep(2, -2, 2))", list(sweep(2, -2, 2)))
    print("list(sweep(2, -2, -2))", list(sweep(2, -2, -2)))
    print("list(sweep_n(0.3, 1.1, 4)", list(sweep_n(0.3, 1.1, 5)))
    print("list(sweep_n(15, -15, 13))", list(sweep_n(15, -15, 13)))
    print("list(sweepLog(0.1, 11e3, 10))", list(sweep_log(0.1, 11e3, 10)))
    print("list(sweep_log(1000, 1, 2))", list(sweep_log(1000, 1, 2)))
    print("list(sweep_log_n(1, 10, 6))", list(sweep_log_n(1, 10, 6)))
    print("list(sweep_log_(10, 1, 5))", list(sweep_log_n(10, 1, 5)))



from spicelib import SpiceEditor
from spicelib.simulators.ltspice_simulator import LTspice

LTspice.create_netlist("./testfiles/Noise.asc")
se = SpiceEditor("./testfiles/Noise.net")

se.set_component_value('R1', 11000)
se.set_component_value('C1', 1.1E-6)
se.set_component_value('V1', 11)

se.run(run_filename="./testfiles/Noise_1.net", timeout=10, simulator=LTspice)



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        test_asc_editor.py
# Purpose:     Tool used validate the LTSpice ASC Files Editor
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

import os
import sys
import unittest

sys.path.append(
    os.path.abspath((os.path.dirname(os.path.abspath(__file__)) + "/../")))  # add project root to lib search path

import spicelib

test_dir = '../examples/testfiles/' if os.path.abspath(os.curdir).endswith('unittests') else './examples/testfiles/'
golden_dir = './golden/' if os.path.abspath(os.curdir).endswith('unittests') else './unittests/golden/'
temp_dir = './temp/' if os.path.abspath(os.curdir).endswith('unittests') else './unittests/temp/'

if not os.path.exists(temp_dir):
    os.mkdir(temp_dir)


def equalFiles(testcase, file1, file2):
    with open(file1, 'r', encoding='cp1252') as f1:
        lines1 = f1.readlines()
    with open(file2, 'r', encoding='cp1252') as f2:
        lines2 = f2.readlines()
    testcase.assertEqual(len(lines1), len(lines2), "Files have different number of lines")
    for i in range(len(lines1)):
        data1 = lines1[i].strip()  # Remove white spaces and line terminators
        data2 = lines2[i].strip()
        if data1.startswith('*') and data2.startswith('*'):
            continue  # Skip comments
        testcase.assertEqual(data1, data2, "Files are not equal")
    

class ASC_Editor_Test(unittest.TestCase):

    def setUp(self):
        self.edt = spicelib.editor.qsch_editor.QschEditor(test_dir + "DC sweep.qsch")

    def test_component_editing(self):
        self.assertEqual(self.edt.get_component_value('R1'), '10K', "Tested R1 Value")  # add assertion here
        self.assertSetEqual(set(self.edt.get_components()), set(('Vin', 'R1', 'R2', 'D1')), "Tested get_components")  # add assertion here
        self.edt.set_component_value('R1', '33K')
        self.edt.save_netlist(temp_dir + 'test_components_output.qsch')
        equalFiles(self, temp_dir + 'test_components_output.qsch', golden_dir + 'test_components_output.qsch')
        self.assertEqual(self.edt.get_component_value('R1'), '33K', "Tested R1 Value")  # add assertion here
        self.edt.remove_component('R1')
        self.edt.save_netlist(temp_dir + 'test_components_output_1.qsch')
        equalFiles(self, temp_dir + 'test_components_output_1.qsch', golden_dir + 'test_components_output_1.qsch')

    def test_parameter_edit(self):
        self.assertEqual(self.edt.get_parameter('TEMP'), '0', "Tested TEMP Parameter")  # add assertion here
        self.edt.set_parameter('TEMP', 25)
        self.assertEqual(self.edt.get_parameter('TEMP'), '25', "Tested TEMP Parameter")  # add assertion here
        self.edt.save_netlist(temp_dir + 'test_parameter_output.qsch')
        equalFiles(self, temp_dir + 'test_parameter_output.qsch', golden_dir + 'test_parameter_output.qsch')
        self.edt.set_parameter('TEMP', 0)  # reset to 0
        self.assertEqual(self.edt.get_parameter('TEMP'), '0.0', "Tested TEMP Parameter")  # add assertion here

    def test_instructions(self):
        self.edt.add_instruction('.ac dec 10 1 100K')
        self.edt.add_instruction('.save V(vout)')
        self.edt.add_instruction('.save I(R1)')
        self.edt.add_instruction('.save I(R2)')
        self.edt.add_instruction('.save I(D1)')
        self.edt.save_netlist(temp_dir + 'test_instructions_output.qsch')
        equalFiles(self, temp_dir + 'test_instructions_output.qsch', golden_dir + 'test_instructions_output.qsch')
        self.edt.remove_instruction('.save I(R1)')
        self.edt.save_netlist(temp_dir + 'test_instructions_output_1.qsch')
        equalFiles(self, temp_dir + 'test_instructions_output_1.qsch', golden_dir + 'test_instructions_output_1.qsch')
        self.edt.remove_Xinstruction(r"\.save\sI\(.*\)")  # removes all .save instructions for currents
        self.edt.save_netlist(temp_dir + 'test_instructions_output_2.qsch')
        equalFiles(self, temp_dir + 'test_instructions_output_2.qsch', golden_dir + 'test_instructions_output_2.qsch')


class QschEditorRotation(unittest.TestCase):

    def test_component_rotations(self):
        self.edt = spicelib.editor.qsch_editor.QschEditor(test_dir + "qsch_rotation.qsch")
        self.edt.save_netlist(temp_dir + 'qsch_rotation.net')
        equalFiles(self, temp_dir + 'qsch_rotation.net', golden_dir + "qsch_rotation.net")


if __name__ == '__main__':
    unittest.main()
